# Yuio project, MIT license.
#
# https://github.com/taminomara/yuio/
#
# You're free to copy this file to your project and edit it for your needs,
# just keep this copyright line please :3

"""
This module implements user-friendly console input and output.

Configuration
-------------

Yuio configures itself upon import using environment variables:

- ``FORCE_NO_COLORS``: disable colored output,
- ``FORCE_COLORS``: enable colored output.

The only thing it doesn't do automatically is wrapping :data:`sys.stdout`
and :data:`sys.stderr` into safe proxies. The :mod:`yuio.app` CLI builder
will do it for you, though, so you don't need to worry about it.

.. autofunction:: setup

To introspect the current state of Yuio's initialization, use the following functions:

.. autofunction:: get_term

.. autofunction:: get_theme

.. autofunction:: wrap_streams

.. autofunction:: restore_streams

.. autofunction:: streams_wrapped

.. autofunction:: orig_stderr

.. autofunction:: orig_stdout


Printing messages
-----------------

To print messages for the user, use these functions:

.. autofunction:: info

.. autofunction:: warning

.. autofunction:: success

.. autofunction:: error

.. autofunction:: error_with_tb

.. autofunction:: heading

.. autofunction:: md

.. autofunction:: br

.. autofunction:: raw


.. _color-tags:

Coloring the output
-------------------

By default, all messages are colored according to their level.

If you need inline colors, you can use special tags in your log messages::

    info('Using the <c code>code</c> tag.')

You can combine multiple colors in the same tag::

    info('<c bold green>Success!</c>')

Only tags that appear in the message itself are processed::

    info('Tags in this message --> %s are printed as-is', '<c color>')

For highlighting inline code, Yuio supports parsing CommonMark's backticks::

    info('Using the `backticks`.')
    info('Using the `` nested `backticks` ``, like they do on GitHub!')

List of all tags that are available by default:

- ``code``, ``note``, ``path``: highlights,
- ``bold``, ``b``, ``dim``, ``d``: font style,
- ``normal``, ``black``, ``red``, ``green``, ``yellow``, ``blue``,
  ``magenta``, ``cyan``, ``white``: colors.


Customizing colors and using themes
-----------------------------------

The :func:`setup` function accepts a :class:`~yuio.theme.Theme` class.
You can subclass it and supply custom colors, see :mod:`yuio.theme`
for more info.


Indicating progress
-------------------

You can use the :class:`Task` class to indicate status and progress
of some task:

.. autoclass:: Task


   .. automethod:: progress

   .. automethod:: progress_size

   .. automethod:: progress_scale

   .. automethod:: iter

   .. automethod:: comment

   .. automethod:: done

   .. automethod:: error

   .. automethod:: subtask


Querying user input
-------------------

If you need to get something from the user, :func:`ask` is the way to do it.

.. autofunction:: ask

.. autofunction:: wait_for_user

You can also prompt the user to edit something with the :func:`edit` function:

.. autofunction:: edit

.. autofunction:: detect_editor

All of these functions throw a error if something goes wrong:

.. autoclass:: UserIoError


Suspending the output
---------------------

You can temporarily disable printing of tasks and messages
using the :class:`SuspendOutput` context manager.

.. autoclass:: SuspendOutput

   .. automethod:: resume

   .. automethod:: info

   .. automethod:: warning

   .. automethod:: success

   .. automethod:: error

   .. automethod:: error_with_tb

   .. automethod:: heading

   .. automethod:: md

   .. automethod:: br

   .. automethod:: raw


Python's `logging` and yuio
---------------------------

If you want to direct messages from the :mod:`logging` to Yuio,
you can add a :class:`Handler`:

.. autoclass:: Handler

"""

from __future__ import annotations

import abc
import atexit
import enum
import functools
import logging
import math
import os
import re
import shutil
import string
import subprocess
import sys
import tempfile
import threading
import time
import traceback
import types
from logging import LogRecord

import yuio.md
import yuio.parse
import yuio.term
import yuio.theme
import yuio.widget
from yuio import _typing as _t

__all__ = [
    "UserIoError",
    "setup",
    "get_term",
    "get_theme",
    "wrap_streams",
    "restore_streams",
    "streams_wrapped",
    "orig_stderr",
    "orig_stdout",
    "info",
    "warning",
    "success",
    "error",
    "error_with_tb",
    "failure",
    "failure_with_tb",
    "heading",
    "md",
    "br",
    "raw",
    "ask",
    "wait_for_user",
    "detect_editor",
    "edit",
    "SuspendOutput",
    "Task",
    "Handler",
]

T = _t.TypeVar("T")
M = _t.TypeVar("M", default=_t.Never)
S = _t.TypeVar("S", default=str)

_ExcInfo: _t.TypeAlias = tuple[
    type[BaseException] | None,
    BaseException | None,
    types.TracebackType | None,
]


_IO_LOCK = threading.Lock()
_IO_MANAGER: _IoManager | None = None
_STREAMS_WRAPPED: bool = False
_ORIG_STDERR: _t.TextIO | None = None
_ORIG_STDOUT: _t.TextIO | None = None


def _manager() -> _IoManager:
    global _IO_MANAGER

    if _IO_MANAGER is None:
        with _IO_LOCK:
            if _IO_MANAGER is None:
                _IO_MANAGER = _IoManager()
    return _IO_MANAGER


class UserIoError(IOError):
    """
    Raised when interaction with user fails.

    """


def setup(
    *,
    term: yuio.term.Term | None = None,
    theme: (
        yuio.theme.Theme | _t.Callable[[yuio.term.Term], yuio.theme.Theme] | None
    ) = None,
    wrap_stdio: bool = True,
):
    """
    Initial setup of the logging facilities.

    :param term:
        terminal that will be used for output.

        If not passed, the global terminal is not re-configured;
        the default is to use a term attached to :data:`sys.stderr`.
    :param theme:
        either a theme that will be used for output, or a theme constructor that takes
        a :class:`~yuio.term.Term` and returns a theme.

        If not passed, the global theme is not re-configured; the default is to use
        :class:`yuio.theme.DefaultTheme` then.
    :param wrap_stdio:
        if set to :data:`True`, wraps :data:`sys.stdout` and :data:`sys.stderr`
        in a special wrapper that ensures better interaction
        with Yuio's progress bars and widgets.

        .. note::

           If you're working with some other library that wraps :data:`sys.stdout`
           and :data:`sys.stderr`, such as colorama_, initialize it before Yuio.

    .. _colorama: https://github.com/tartley/colorama

    """

    global _IO_MANAGER

    if not (manager := _IO_MANAGER):
        with _IO_LOCK:
            if not (manager := _IO_MANAGER):
                _IO_MANAGER = _IoManager(term, theme)
    if manager is not None:
        manager.setup(term, theme)

    if wrap_stdio:
        wrap_streams()


def get_term() -> yuio.term.Term:
    """
    Get the global instance of :class:`~yuio.term.Term` that is used
    with :mod:`yuio.io`.

    If global setup wasn't performed, this function implicitly performs it.

    """

    return _manager().term


def get_theme() -> yuio.theme.Theme:
    """
    Get the global instance of :class:`~yuio.theme.Theme`
    that is used with :mod:`yuio.io`.

    If global setup wasn't performed, this function implicitly performs it.

    """

    return _manager().theme


def wrap_streams():
    """
    Wrap :data:`sys.stdout` and :data:`sys.stderr` so that they honor
    Yuio tasks and widgets.

    .. note::

        If you're working with some other library that wraps :data:`sys.stdout`
        and :data:`sys.stderr`, such as colorama_, initialize it before Yuio.

    See :func:`setup`.

    .. _colorama: https://github.com/tartley/colorama

    """

    global _STREAMS_WRAPPED, _ORIG_STDOUT, _ORIG_STDERR

    with _IO_LOCK:
        if _STREAMS_WRAPPED:
            return

        if yuio.term._is_interactive_output(sys.stdout):
            _ORIG_STDOUT, sys.stdout = sys.stdout, _WrappedOutput(sys.stdout)
        if yuio.term._is_interactive_output(sys.stderr):
            _ORIG_STDERR, sys.stderr = sys.stderr, _WrappedOutput(sys.stderr)
        _STREAMS_WRAPPED = True

        atexit.register(restore_streams)


def restore_streams():
    """
    Restore wrapped streams.

    See :func:`wrap_streams` and :func:`setup`.

    """

    global _STREAMS_WRAPPED

    with _IO_LOCK:
        if not _STREAMS_WRAPPED:
            return

        if _ORIG_STDOUT is not None:
            sys.stdout = _ORIG_STDOUT
        if _ORIG_STDERR is not None:
            sys.stderr = _ORIG_STDERR
        _STREAMS_WRAPPED = False


def streams_wrapped() -> bool:
    """
    Check if :data:`sys.stdout` and :data:`sys.stderr` are wrapped.
    See :func:`setup`.

    """

    return _STREAMS_WRAPPED


def orig_stderr() -> _t.TextIO:
    """
    Return the original :data:`sys.stderr` before wrapping."""

    return _ORIG_STDERR or sys.stderr


def orig_stdout() -> _t.TextIO:
    """
    Return the original :data:`sys.stdout` before wrapping."""

    return _ORIG_STDOUT or sys.stdout


def info(msg: str, /, *args, **kwargs):
    """
    Print an info message."""

    _manager().print_msg(msg, args, "info", **kwargs)


def warning(msg: str, /, *args, **kwargs):
    """
    Print a warning message."""

    _manager().print_msg(msg, args, "warning", **kwargs)


def success(msg: str, /, *args, **kwargs):
    """
    Print a success message."""

    _manager().print_msg(msg, args, "success", **kwargs)


def error(msg: str, /, *args, **kwargs):
    """
    Print an error message."""

    _manager().print_msg(msg, args, "error", **kwargs)


def error_with_tb(
    msg: str, /, *args, exc_info: _ExcInfo | bool | None = True, **kwargs
):
    """
    Print an error message and capture the current exception.

    Call this function in the ``except`` clause of a ``try`` block
    or in an ``__exit__`` function of a context manager to attach
    current exception details to the log message.

    :param exc_info:
        either a boolean indicating that the current exception
        should be captured (default is :data:`True`), or a tuple
        of three elements, as returned by :func:`sys.exc_info`.

    """

    _manager().print_msg(msg, args, "error", exc_info=exc_info, **kwargs)


def failure(msg: str, /, *args, **kwargs):
    """
    Print a failure message."""

    _manager().print_msg(msg, args, "failure", **kwargs)


def failure_with_tb(
    msg: str, /, *args, exc_info: _ExcInfo | bool | None = True, **kwargs
):
    """
    Print a failure message and capture the current exception.

    Call this function in the ``except`` clause of a ``try`` block
    or in an ``__exit__`` function of a context manager to attach
    current exception details to the log message.

    :param exc_info:
        either a boolean indicating that the current exception
        should be captured (default is :data:`True`), or a tuple
        of three elements, as returned by :func:`sys.exc_info`.

    """

    _manager().print_msg(msg, args, "failure", exc_info=exc_info, **kwargs)


def heading(msg: str, /, *args, **kwargs):
    """
    Print a heading message."""

    kwargs.setdefault("heading", True)
    _manager().print_msg(msg, args, "heading/1", **kwargs)


def md(msg: str, /, *args, **kwargs):
    """
    Print a markdown-formatted text.

    Yuio supports all CommonMark block markup except tables. Inline markup is limited
    to backticks and color tags.

    See :mod:`yuio.md` for more info.

    """

    _manager().print_md(msg, args, **kwargs)


def br(**kwargs):
    """
    Print an empty string."""

    _manager().print_direct("\n", **kwargs)


def raw(msg: yuio.term.ColorizedString, /, **kwargs):
    """
    Print a :class:`~yuio.term.ColorizedString`.

    This is a bridge between :mod:`yuio.io` and lower-level
    modules like :mod:`yuio.term`.

    In most cases, you won't need this function. The only exception
    is when you need to build a :class:`~yuio.term.ColorizedString`
    yourself.

    """

    _manager().print_raw(msg, **kwargs)


class _AskWidget(yuio.widget.Widget[T], _t.Generic[T]):
    _layout: yuio.widget.VerticalLayout[T]

    def __init__(
        self, prompt: yuio.term.ColorizedString, widget: yuio.widget.Widget[T]
    ):
        self._prompt = yuio.widget.Text(prompt)
        self._error_msg: str | None = None
        self._inner = widget

    def event(self, e: yuio.widget.KeyboardEvent, /) -> yuio.widget.Result[T] | None:
        try:
            result = self._inner.event(e)
        except yuio.parse.ParsingError as err:
            self._error_msg = f"Error: {err}."
        else:
            self._error_msg = None
            return result

    def layout(self, rc: yuio.widget.RenderContext, /) -> tuple[int, int]:
        builder = (
            yuio.widget.VerticalLayoutBuilder()
            .add(self._prompt)
            .add(self._inner, receive_events=True)
        )
        if self._error_msg is not None:
            rc.bell()
            error_text = yuio.term.ColorizedString(
                [
                    rc.theme.get_color("msg/decoration:error"),
                    rc.theme.msg_decorations.get("error", "▲ "),
                    rc.theme.get_color("msg/text:error"),
                    self._error_msg,
                    yuio.term.Color.NONE,
                ]
            )
            builder = builder.add(yuio.widget.Text(error_text))

        self._layout = builder.build()
        return self._layout.layout(rc)

    def draw(self, rc: yuio.widget.RenderContext, /):
        self._layout.draw(rc)

    @property
    def help_data(self) -> yuio.widget.WidgetHelp:
        return self._inner.help_data


class _AskMeta(type):
    __hint = None

    @_t.overload
    def __call__(  # pyright: ignore[reportInconsistentOverload]
        cls: type[ask[S]],
        msg: str,
        /,
        *args,
        default: M | yuio.Missing = yuio.MISSING,
        parser: yuio.parse.Parser[S] | None = None,
        input_description: str | None = None,
        default_description: str | None = None,
    ) -> S | M: ...

    def __call__(cls, *args, **kwargs):
        if "parser" not in kwargs:
            hint = cls.__hint
            if hint is None:
                hint = str
            kwargs["parser"] = yuio.parse.from_type_hint(hint)
        return _ask(*args, **kwargs)

    def __getitem(cls, ty):
        return _AskMeta("ask", (), {"_AskMeta__hint": ty})

    # A dirty hack to hide `__getitem__` from type checkers. `ask` should look like
    # an ordinary class with overloaded `__new__` for the magic to work.
    locals()["__getitem__"] = __getitem

    def __repr__(cls) -> str:
        if cls.__hint is None:
            return cls.__name__
        else:
            return f"{cls.__name__}[{_t.type_repr(cls.__hint)}]"


@_t.final
class ask(_t.Generic[S], metaclass=_AskMeta):
    """
    Ask user to provide an input, parse it and return a value.

    If `stdin` is not readable, return default if one is present,
    or raise a :class:`UserIoError`.

    .. vhs:: /_tapes/questions.tape
        :alt: Demonstration of the `ask` function.
        :scale: 40%

    :func:`ask` accepts generic parameters, which determine how input is parsed.
    For example, if you're asking for an enum element,
    Yuio will show user a choice widget.

    You can also supply a custom :class:`~yuio.parse.Parser`,
    which will determine the widget that is displayed to the user,
    the way auto completions work, etc.

    Example:

    .. code-block:: python

        class Level(enum.Enum):
            WARNING = "Warning",
            INFO = "Info",
            DEBUG = "Debug",

        answer = ask[Level]('Choose a logging level', default=Level.INFO)

    :param msg:
        prompt to display to user.
    :param args:
        arguments for prompt formatting.
    :param parser:
        parser to use to parse user input. See :mod:`yuio.parse` for more info.
    :param default:
        default value to return if user input is empty.
    :param input_description:
        description of the expected input, like ``"yes/no"`` for boolean
        inputs.
    :param default_description:
        description of the `default` value.

    """


def _ask(
    msg: str,
    /,
    *args,
    parser: yuio.parse.Parser[_t.Any],
    default: _t.Any = yuio.MISSING,
    input_description: str | None = None,
    default_description: str | None = None,
) -> _t.Any:
    manager = _manager()

    term, formatter, theme = manager.term, manager.formatter, manager.theme

    if not term.istream.readable():
        if default is not yuio.MISSING:
            return default
        else:
            raise UserIoError("can't interact with user in non-interactive environment")

    if default is None and not yuio.parse._is_optional_parser(parser):
        parser = yuio.parse.Optional(parser)

    msg = msg.rstrip()
    if msg.endswith(":"):
        needs_colon = True
        msg = msg[:-1]
    else:
        needs_colon = msg and not msg[-1] in string.punctuation

    prompt = formatter.colorize(msg, default_color="msg/text:question")
    if args:
        prompt = prompt % args

    if not input_description:
        input_description = parser.describe()

    if default is not yuio.MISSING:
        if default_description is None:
            try:
                default_description = parser.describe_value(default)
            except TypeError:
                pass
        if default_description is None:
            default_description = str(default)

    if get_term().is_fully_interactive:
        # Use widget.

        if needs_colon:
            prompt += formatter.colorize(":", default_color="msg/text:question")

        widget = _AskWidget(
            prompt, parser.widget(default, input_description, default_description)
        )
        with SuspendOutput() as s:
            try:
                result = widget.run(term, theme)
            except (OSError, EOFError) as e:
                raise UserIoError("unexpected end of input") from e

            if result is yuio.MISSING:
                result = default

            try:
                result_desc = parser.describe_value_or_def(result)
            except TypeError:
                result_desc = str(result)

            confirmation = prompt + (formatter.colorize(" `%s`\n") % result_desc)

            s.raw(confirmation)
            return result
    else:
        # Use raw input.

        if input_description:
            prompt += (
                formatter.colorize(" (%s)", default_color="msg/text:question")
                % input_description
            )
        if default_description:
            prompt += (
                formatter.colorize(" [`%s`]", default_color="msg/text:question")
                % default_description
            )
        prompt += formatter.colorize(
            ": " if needs_colon else " ", default_color="msg/text:question"
        )
        with SuspendOutput() as s:
            while True:
                try:
                    s.raw(prompt)
                    answer = term.istream.readline().strip()
                except (OSError, EOFError) as e:
                    raise UserIoError("unexpected end of input") from None
                if not answer and default is not yuio.MISSING:
                    return default
                elif not answer:
                    s.error("Input is required.")
                else:
                    try:
                        return parser.parse(answer)
                    except yuio.parse.ParsingError as e:
                        s.error(f"Error: %s.", e)


class _WaitForUserWidget(yuio.widget.Widget[None]):
    def __init__(self, prompt: yuio.term.ColorizedString):
        self._prompt = yuio.widget.Text(prompt)

    def layout(self, rc: yuio.widget.RenderContext, /) -> tuple[int, int]:
        return self._prompt.layout(rc)

    def draw(self, rc: yuio.widget.RenderContext, /):
        return self._prompt.draw(rc)

    @yuio.widget.bind(yuio.widget.Key.ENTER)
    @yuio.widget.bind(yuio.widget.Key.ESCAPE)
    @yuio.widget.bind("d", ctrl=True)
    @yuio.widget.bind(" ")
    def exit(self):
        return yuio.widget.Result(None)


def wait_for_user(
    msg: str = "Press <c note>enter</c> to continue",
    /,
    *args,
):
    """
    A simple function to wait for user to press enter.

    If `stdin` is not readable, does not do anything.

    """

    manager = _manager()

    term, formatter, theme = manager.term, manager.formatter, manager.theme

    if not term.istream.readable():
        return

    if msg and not msg[-1].isspace():
        msg += " "

    formatter = _manager().formatter
    term = get_term()

    prompt = formatter.colorize(msg, default_color="msg/text:question")
    if args:
        prompt %= args

    with SuspendOutput() as s:
        try:
            if term.is_fully_interactive:
                _WaitForUserWidget(prompt).run(term, theme)
            else:
                s.raw(prompt)
                term.istream.readline()
        except (OSError, EOFError):
            return


def detect_editor() -> str | None:
    """
    Detect the user's preferred editor.

    This function checks the ``EDITOR`` environment variable.
    If it's not found, it checks whether ``nano`` or ``vi``
    are available. Otherwise, it returns `None`.

    """

    if editor := os.environ.get("EDITOR"):
        return editor
    elif editor := shutil.which("nano"):
        return editor
    elif editor := shutil.which("vi"):
        return editor
    elif editor := shutil.which("notepad.exe"):
        return editor
    else:
        return None


def edit(
    text: str,
    /,
    *,
    comment_marker: str | None = "#",
    editor: str | None = None,
    file_ext: str | None = None,
) -> str:
    """
    Ask user to edit some text.

    This function creates a temporary file with the given text
    and opens it in an editor. After editing is done, it strips away
    all lines that start with `comment_marker`, if one is given.

    If editor is not available or returns a non-zero exit code,
    a :class:`UserIoError` is raised.

    If launched in a non-interactive environment, returns the text
    unedited (comments are still removed, though).

    """

    if _manager().term.is_fully_interactive:
        if editor is None:
            editor = detect_editor()

        if editor is None:
            raise UserIoError(
                "can't detect an editor, ensure that the $EDITOR "
                "environment variable contains "
                "a correct path to an editor executable"
            )

        fd, filepath = tempfile.mkstemp(text=True, suffix=file_ext)
        try:
            with open(fd, "w") as file:
                file.write(text)

            try:
                with SuspendOutput():
                    res = subprocess.run(f'{editor} "{filepath}"', shell=True)
            except FileNotFoundError:
                raise UserIoError(
                    "can't use this editor, ensure that the $EDITOR "
                    "environment variable contains "
                    "a correct path to an editor executable"
                )

            if res.returncode != 0:
                raise UserIoError("editing failed")

            if not os.path.exists(filepath):
                text = ""
            else:
                with open(filepath) as file:
                    text = file.read()
        finally:
            try:
                os.remove(filepath)
            except OSError:
                pass

    if comment_marker is not None:
        text = re.sub(
            r"^\s*" + re.escape(comment_marker) + r".*(\n|$)",
            "",
            text,
            flags=re.MULTILINE,
        )

    return text


class SuspendOutput:
    """
    A context manager for pausing output.

    This is handy for when you need to take control over the output stream.
    For example, the :func:`ask` function uses this class internally.

    This context manager also suspends all prints that go to :data:`sys.stdout`
    and :data:`sys.stderr` if they were wrapped (see :func:`setup`).
    To print through them, use :func:`orig_stderr` and :func:`orig_stdout`.

    """

    def __init__(self):
        self._resumed = False
        _manager().suspend()

    def resume(self):
        """
        Manually resume the logging process."""

        if not self._resumed:
            _manager().resume()
            self._resumed = True

    @staticmethod
    def info(msg: str, /, *args, **kwargs):
        """
        Log an :func:`info` message, ignore suspended status."""

        kwargs.setdefault("ignore_suspended", True)
        info(msg, *args, **kwargs)

    @staticmethod
    def warning(msg: str, /, *args, **kwargs):
        """
        Log a :func:`warning` message, ignore suspended status."""

        kwargs.setdefault("ignore_suspended", True)
        warning(msg, *args, **kwargs)

    @staticmethod
    def success(msg: str, /, *args, **kwargs):
        """
        Log a :func:`success` message, ignore suspended status."""

        kwargs.setdefault("ignore_suspended", True)
        success(msg, *args, **kwargs)

    @staticmethod
    def error(msg: str, /, *args, **kwargs):
        """
        Log an :func:`error` message, ignore suspended status."""

        kwargs.setdefault("ignore_suspended", True)
        error(msg, *args, **kwargs)

    @staticmethod
    def error_with_tb(msg: str, /, *args, **kwargs):
        """
        Log an :func:`error_with_tb` message, ignore suspended status."""

        kwargs.setdefault("ignore_suspended", True)
        error_with_tb(msg, *args, **kwargs)

    @staticmethod
    def failure(msg: str, /, *args, **kwargs):
        """
        Log an :func:`failure` message, ignore suspended status."""

        kwargs.setdefault("ignore_suspended", True)
        failure(msg, *args, **kwargs)

    @staticmethod
    def failure_with_tb(msg: str, /, *args, **kwargs):
        """
        Log an :func:`failure_with_tb` message, ignore suspended status."""

        kwargs.setdefault("ignore_suspended", True)
        failure_with_tb(msg, *args, **kwargs)

    @staticmethod
    def heading(msg: str, /, *args, **kwargs):
        """
        Log a :func:`heading` message, ignore suspended status."""

        kwargs.setdefault("ignore_suspended", True)
        heading(msg, *args, **kwargs)

    @staticmethod
    def md(msg: str, /, *args, **kwargs):
        """
        Log a markdown-formatted text."""

        kwargs.setdefault("ignore_suspended", True)
        md(msg, *args, **kwargs)

    @staticmethod
    def br(**kwargs):
        """
        Log an empty string."""

        kwargs.setdefault("ignore_suspended", True)
        br(**kwargs)

    @staticmethod
    def raw(msg: yuio.term.ColorizedString, **kwargs):
        """
        Log a :class:`~yuio.term.ColorizedString`."""

        kwargs.setdefault("ignore_suspended", True)
        raw(msg, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.resume()


class _IterTask(_t.Generic[T]):
    def __init__(
        self, collection: _t.Collection[T], task: Task, unit: str, ndigits: int
    ):
        self._iter = iter(collection)
        self._task = task
        self._unit = unit
        self._ndigits = ndigits

        self._i = 0
        self._len = len(collection)

    def __next__(self) -> T:
        self._task.progress(self._i, self._len, unit=self._unit, ndigits=self._ndigits)
        if self._i < self._len:
            self._i += 1
        return self._iter.__next__()

    def __iter__(self) -> _IterTask[T]:
        return self


class Task:
    """
    A class for indicating progress of some task.

    You can have multiple tasks at the same time,
    create subtasks, set task's progress or add a comment about
    what's currently being done within a task.

    .. vhs:: /_tapes/tasks_multithreaded.tape
       :alt: Demonstration of the `Task` class.
       :scale: 40%

    This class can be used as a context manager::

        with Task('Processing input') as t:
            ...
            t.progress(0.3)
            ...

    """

    class _Status(enum.Enum):
        RUNNING = "running"
        DONE = "done"
        ERROR = "error"

    def __init__(
        self,
        msg: str,
        /,
        *args,
        _parent: Task | None = None,
        comment: str | None = None,
    ):
        # Task properties should not be written to directly.
        # Instead, task should be sent to a handler for modification.
        # This ensures thread safety, because handler has a lock.
        # See handler's implementation details.

        self._msg: str = msg
        self._args: tuple[object, ...] = args
        self._comment: str | None = comment
        self._comment_args: tuple[object, ...] | None = None
        self._status: Task._Status = Task._Status.RUNNING
        self._progress: float | None = None
        self._progress_done: str | None = None
        self._progress_total: str | None = None
        self._subtasks: list[Task] = []

        self._cached_msg: yuio.term.ColorizedString | None = None
        self._cached_comment: yuio.term.ColorizedString | None = None

        if _parent is None:
            _manager().start_task(self)
        else:
            _manager().start_subtask(_parent, self)

    @_t.overload
    def progress(self, progress: float | None, /, *, ndigits: int = 2): ...

    @_t.overload
    def progress(
        self,
        done: float | int,
        total: float | int,
        /,
        *,
        unit: str = "",
        ndigits: int = 0,
    ): ...

    def progress(
        self,
        *args: float | int | None,
        unit: str = "",
        ndigits: int | None = None,
    ):
        """
        Indicate progress of this task.

        If given one argument, it is treated as percentage between `0` and `1`.

        If given two arguments, they are treated as amount of finished work,
        and a total amount of work. In this case, optional argument `unit`
        can be used to indicate units, in which amount is calculated::

            with Task("Loading cargo") as task:
                task.progress(110, 150, unit="Kg")

        This will print the following:

        .. code-block:: text

           ■■■■■■■■■■■□□□□ Loading cargo - 110/150Kg

        If given a single :data:`None`, reset task progress.

        """

        progress = None

        if len(args) == 1:
            progress = done = args[0]
            total = None
            if ndigits is None:
                ndigits = 2
        elif len(args) == 2:
            done, total = args
            if ndigits is None:
                ndigits = (
                    2 if isinstance(done, float) or isinstance(total, float) else 0
                )
        else:
            raise ValueError(
                f"Task.progress() takes between one and two arguments "
                f"({len(args)} given)"
            )

        if done is None:
            _manager().set_task_progress(self, None, None, None)
            return

        if len(args) == 1:
            done *= 100
            unit = "%"

        done_str = "%.*f" % (ndigits, done)
        if total is None:
            _manager().set_task_progress(self, progress, done_str + unit, None)
        else:
            total_str = "%.*f" % (ndigits, total)
            progress = done / total if total else 0
            _manager().set_task_progress(self, progress, done_str, total_str + unit)

    def progress_size(
        self,
        done: float | int,
        total: float | int,
        /,
        *,
        ndigits: int = 2,
    ):
        """
        Indicate progress of this task using human-readable 1024-based size units.

        Example::

            with Task("Downloading a file") as task:
                task.progress_size(31.05 * 2**20, 150 * 2**20)

        This will print:

        .. code-block:: text

           ■■■□□□□□□□□□□□□ Downloading a file - 31.05/150.00M

        """

        progress = done / total
        done, done_unit = self._size(done)
        total, total_unit = self._size(total)

        if done_unit == total_unit:
            done_unit = ""

        _manager().set_task_progress(
            self,
            progress,
            "%.*f%s" % (ndigits, done, done_unit),
            "%.*f%s" % (ndigits, total, total_unit),
        )

    @staticmethod
    def _size(n):
        for unit in "BKMGT":
            if n < 1024:
                return n, unit
            n /= 1024
        return n, "P"

    def progress_scale(
        self,
        done: float | int,
        total: float | int,
        /,
        *,
        unit: str = "",
        ndigits: int = 2,
    ):
        """
        Indicate progress of this task while scaling numbers in accordance
        with SI system.

        Example::

            with Task("Charging a capacitor") as task:
                task.progress_scale(889.25E-3, 1, unit="V")


        This will print:

        .. code-block:: text

           ■■■■■■■■■■■■■□□ Charging a capacitor - 889.25mV/1.00V

        """

        progress = done / total
        done, done_unit = self._unit(done)
        total, total_unit = self._unit(total)

        if unit:
            done_unit += unit
            total_unit += unit

        _manager().set_task_progress(
            self,
            progress,
            "%.*f%s" % (ndigits, done, done_unit),
            "%.*f%s" % (ndigits, total, total_unit),
        )

    @staticmethod
    def _unit(n: float) -> tuple[float, str]:
        if math.fabs(n) < 1e-33:
            return 0, ""
        magnitude = max(-8, min(8, int(math.log10(math.fabs(n)) // 3)))
        if magnitude < 0:
            return n * 10 ** -(3 * magnitude), "mµnpfazy"[-magnitude - 1]
        elif magnitude > 0:
            return n / 10 ** (3 * magnitude), "KMGTPEZY"[magnitude - 1]
        else:
            return n, ""

    def iter(
        self,
        collection: _t.Collection[T],
        /,
        *,
        unit: str = "",
        ndigits: int = 0,
    ) -> _t.Iterable[T]:
        """
        Helper for updating progress automatically
        while iterating over a collection.

        For example::

            with Task('Fetching data') as t:
                for url in t.iter(urls):
                    ...

        This will output the following:

        .. code-block:: text

           ■■■■■□□□□□□□□□□ Fetching data - 1/3

        """

        return _IterTask(collection, self, unit, ndigits)

    def comment(self, comment: str | None, /, *args):
        """
        Set a comment for a task.

        Comment is displayed after the progress.

        For example::

            with Task('Fetching data') as t:
                for url in urls:
                    t.comment(url)
                    ...

        This will output the following:

        .. code-block:: text

           ⣿ Fetching data - https://google.com

        """

        _manager().set_task_comment(self, comment, args)

    def done(self):
        """
        Indicate that this task has finished successfully."""

        _manager().finish_task(self, Task._Status.DONE)

    def error(self):
        """
        Indicate that this task has finished with an error."""

        _manager().finish_task(self, Task._Status.ERROR)

    def subtask(self, msg: str, /, *args) -> Task:
        """
        Create a subtask within this task."""

        return Task(msg, *args, _parent=self)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self.done()
        else:
            self.error()


class Handler(logging.Handler):
    """
    A handler that redirects all log messages to yuio.

    """

    def emit(self, record: LogRecord):
        _manager().print_rec(record)


class _IoManager(abc.ABC):
    term: yuio.term.Term
    theme: yuio.theme.Theme

    def __init__(
        self,
        term: yuio.term.Term | None = None,
        theme: (
            yuio.theme.Theme | _t.Callable[[yuio.term.Term], yuio.theme.Theme] | None
        ) = None,
        enable_bg_updates: bool = True,
    ):
        self.term = term or yuio.term.get_term_from_stream(orig_stderr(), sys.stdin)
        if theme is None:
            self.theme = yuio.theme.load(self.term)
        elif isinstance(theme, yuio.theme.Theme):
            self.theme = theme
        else:
            self.theme = theme(self.term)
        self.formatter = yuio.md.MdFormatter(self.theme)
        self._rc = yuio.widget.RenderContext(self.term, self.theme)

        self._suspended: int = 0
        self._suspended_lines: list[tuple[list[str], _t.TextIO]] = []

        self._tasks: list[Task] = []
        self._tasks_printed = 0
        self._spinner_state = 0
        self._needs_update = False
        self._last_update_time_us = 0
        self._printed_some_lines = False

        self._renders = 0

        self._stop = False
        self._stop_condition = threading.Condition(_IO_LOCK)
        self._thread: threading.Thread | None = None

        self._enable_bg_updates = enable_bg_updates
        if enable_bg_updates:
            self._thread = threading.Thread(
                target=self._bg_update, name="yuio_io_task_refresh", daemon=True
            )
            self._thread.start()

            atexit.register(self.stop)

    def setup(
        self,
        term: yuio.term.Term | None = None,
        theme: (
            yuio.theme.Theme | _t.Callable[[yuio.term.Term], yuio.theme.Theme] | None
        ) = None,
    ):
        self._clear_tasks()

        if term is not None:
            self.term = term
        if theme is not None:
            if not isinstance(theme, yuio.theme.Theme):
                theme = theme(self.term)
            self.theme = self.formatter.theme = theme

        self._rc = yuio.widget.RenderContext(self.term, self.theme)
        self.__dict__.pop("update_rate_us", None)
        self._update_tasks()

    @functools.cached_property
    def update_rate_us(self) -> int:
        update_rate_ms = max(self.theme.spinner_update_rate_ms, 1)
        while update_rate_ms < 50:
            update_rate_ms *= 2
        while update_rate_ms > 100:
            update_rate_ms /= 2
        return int(update_rate_ms * 1000)

    @property
    def spinner_update_rate_us(self) -> int:
        return self.theme.spinner_update_rate_ms * 1000

    def _bg_update(self):
        while True:
            try:
                with _IO_LOCK:
                    while True:
                        update_rate_us = self.update_rate_us
                        now_us = time.monotonic_ns() // 1_000
                        sleep_us = update_rate_us - now_us % update_rate_us

                        if self._stop_condition.wait_for(
                            lambda: self._stop, timeout=sleep_us / 1_000_000
                        ):
                            return

                        self._show_tasks()
            except Exception:
                yuio._logger.critical("exception in bg updater", exc_info=True)

    def stop(self):
        with _IO_LOCK:
            atexit.unregister(self.stop)

            self._stop = True
            self._stop_condition.notify()
            self._show_tasks(immediate_render=True)

        if self._thread:
            self._thread.join()

    def print_msg(
        self,
        msg: str,
        args: tuple[object, ...] | None,
        tag: str,
        /,
        *,
        color: str | None = None,
        exc_info: _ExcInfo | bool | None = None,
        ignore_suspended: bool = False,
        heading: bool = False,
    ):
        with _IO_LOCK:
            col_msg = self._format_msg(
                msg, args, color or tag, exc_info=exc_info, heading=heading
            )
            self._emit_lines(col_msg.process_colors(self.term), None, ignore_suspended)

    def print_md(
        self,
        msg: str,
        args: tuple[object, ...] | None,
        /,
        *,
        ignore_suspended: bool = False,
    ):
        with _IO_LOCK:
            col_md = self._format_md(msg, args)
            self._emit_lines(col_md.process_colors(self.term), None, ignore_suspended)

    def print_rec(
        self,
        record: logging.LogRecord,
    ):
        with _IO_LOCK:
            col_rec = self._format_rec(record)
            self._emit_lines(col_rec.process_colors(self.term))

    def print_raw(
        self,
        msg: yuio.term.ColorizedString,
        /,
        *,
        ignore_suspended: bool = False,
    ):
        with _IO_LOCK:
            self._emit_lines(msg.process_colors(self.term), None, ignore_suspended)

    def print_direct(
        self,
        msg: str,
        stream: _t.TextIO | None = None,
        /,
        *,
        ignore_suspended: bool = False,
    ):
        with _IO_LOCK:
            self._emit_lines([msg], stream, ignore_suspended)

    def print_direct_lines(
        self,
        lines: _t.Iterable[str],
        stream: _t.TextIO | None = None,
        /,
        *,
        ignore_suspended: bool = False,
    ):
        with _IO_LOCK:
            self._emit_lines(lines, stream, ignore_suspended)

    def start_task(self, task: Task):
        with _IO_LOCK:
            self._start_task(task)

    def start_subtask(self, parent: Task, task: Task):
        with _IO_LOCK:
            self._start_subtask(parent, task)

    def finish_task(self, task: Task, status: Task._Status):
        with _IO_LOCK:
            self._finish_task(task, status)

    def set_task_progress(
        self,
        task: Task,
        progress: float | None,
        done: str | None,
        total: str | None,
    ):
        with _IO_LOCK:
            task._progress = progress
            task._progress_done = done
            task._progress_total = total
            self._update_tasks()

    def set_task_comment(self, task: Task, comment: str | None, args):
        with _IO_LOCK:
            task._comment = comment
            task._comment_args = args
            task._cached_comment = None
            self._update_tasks()

    def suspend(self):
        with _IO_LOCK:
            self._suspend()

    def resume(self):
        with _IO_LOCK:
            self._resume()

    # Implementation.
    # These functions are always called under a lock.

    def _emit_lines(
        self,
        lines: _t.Iterable[str],
        stream: _t.TextIO | None = None,
        ignore_suspended: bool = False,
    ):
        stream = stream or self.term.ostream
        if self._suspended and not ignore_suspended:
            self._suspended_lines.append((list(lines), stream))
        else:
            self._clear_tasks()
            stream.writelines(lines)
            self._update_tasks(immediate_render=True)
            stream.flush()

        self._printed_some_lines = True

    def _suspend(self):
        self._suspended += 1

        if self._suspended == 1:
            self._clear_tasks()

    def _resume(self):
        self._suspended -= 1

        if self._suspended == 0:
            for lines, stream in self._suspended_lines:
                stream.writelines(lines)
            if self._suspended_lines:
                self._printed_some_lines = True
            self._suspended_lines.clear()

            self._update_tasks()

        if self._suspended < 0:
            yuio._logger.debug("unequal number of suspends and resumes")
            self._suspended = 0

    def _start_task(self, task: Task):
        if self.term.can_move_cursor:
            self._tasks.append(task)
            self._update_tasks()
        else:
            self._emit_lines(self._format_task(task).process_colors(self.term))

    def _start_subtask(self, parent: Task, task: Task):
        if self.term.can_move_cursor:
            parent._subtasks.append(task)
            self._update_tasks()
        else:
            self._emit_lines(self._format_task(task).process_colors(self.term))

    def _finish_task(self, task: Task, status: Task._Status):
        if task._status != Task._Status.RUNNING:
            yuio._logger.debug("trying to change status of an already stopped task")
            return

        task._status = status

        if self.term.can_move_cursor:
            if task in self._tasks:
                self._tasks.remove(task)
                self._emit_lines(self._format_task(task).process_colors(self.term))
            else:
                self._update_tasks()
        else:
            self._emit_lines(self._format_task(task).process_colors(self.term))

    def _clear_tasks(self):
        if self.term.can_move_cursor and self._tasks_printed:
            self._rc.finalize()
            self._tasks_printed = 0

    def _update_tasks(self, immediate_render: bool = False):
        self._needs_update = True
        if immediate_render or not self._enable_bg_updates:
            self._show_tasks(immediate_render)

    def _show_tasks(self, immediate_render: bool = False):
        if (
            self.term.can_move_cursor
            and not self._suspended
            and (self._tasks or self._tasks_printed)
        ):
            now_us = time.monotonic_ns() // 1000
            now_us -= now_us % self.update_rate_us

            if not immediate_render and self._enable_bg_updates:
                next_update_us = self._last_update_time_us + self.update_rate_us
                if now_us < next_update_us:
                    # Hard-limit update rate by `update_rate_ms`.
                    return
                next_spinner_update_us = (
                    self._last_update_time_us + self.spinner_update_rate_us
                )
                if not self._needs_update and now_us < next_spinner_update_us:
                    # Tasks didn't change, and spinner state didn't change either,
                    # so we can skip this update.
                    return

            self._last_update_time_us = now_us
            self._spinner_state = now_us // self.spinner_update_rate_us
            self._tasks_printed = 0
            self._needs_update = False

            self._rc.prepare()
            self.formatter.width = self._rc.width
            for task in self._tasks:
                self._draw_task(task, 0)
            self._renders += 1
            self._rc.set_final_pos(0, self._tasks_printed)
            self._rc.render()

    def _format_msg(
        self,
        msg: str,
        args: tuple[object, ...] | None,
        tag: str,
        /,
        *,
        exc_info: _ExcInfo | bool | None = None,
        heading: bool = False,
    ) -> yuio.term.ColorizedString:
        decoration = self.theme.msg_decorations.get(tag, "")
        if decoration:
            first_line_indent = yuio.term.ColorizedString(
                [self.theme.get_color(f"msg/decoration:{tag}"), decoration]
            )
            continuation_indent = " " * first_line_indent.width
        else:
            first_line_indent = ""
            continuation_indent = ""

        res = yuio.term.ColorizedString()

        if heading and self._printed_some_lines:
            res += "\n"

        col_msg = self.formatter.colorize(msg, default_color=f"msg/text:{tag}")
        if args:
            col_msg %= args
        for line in col_msg.wrap(
            self.formatter.width,
            first_line_indent=first_line_indent,
            continuation_indent=continuation_indent,
        ):
            res += line
            res += "\n"

        if exc_info is True:
            exc_info = sys.exc_info()
        elif exc_info is False or exc_info is None:
            exc_info = None
        elif isinstance(exc_info, BaseException):
            exc_info = (type(exc_info), exc_info, exc_info.__traceback__)
        elif not isinstance(exc_info, tuple) or len(exc_info) != 3:
            raise ValueError(f"invalid exc_info {exc_info!r}")
        if exc_info is not None and exc_info != (None, None, None):
            tb = "".join(traceback.format_exception(*exc_info))
            res += self._format_tb(tb, "  ")

        if heading:
            res += "\n"

        res += yuio.term.Color.NONE

        return res

    def _format_md(self, md: str, args) -> yuio.term.ColorizedString:
        res = yuio.term.ColorizedString()
        for line in self.formatter.format(md):
            res += line
            res += "\n"
        if args:
            res %= args
        return res

    def _format_rec(self, record: logging.LogRecord) -> yuio.term.ColorizedString:
        res = yuio.term.ColorizedString()

        asctime = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(record.created))

        logger = record.name
        level = record.levelname
        message = record.getMessage()

        ctx = record.levelname.lower()

        res += self.theme.get_color(f"log/asctime:{ctx}")
        res += asctime
        res += " "

        res += self.theme.get_color(f"log/logger:{ctx}")
        res += logger
        res += " "

        res += self.theme.get_color(f"log/level:{ctx}")
        res += level
        res += " "

        res += self.theme.get_color(f"log/message:{ctx}")
        res += message
        res += "\n"

        if record.exc_info:
            if not record.exc_text:
                record.exc_text = "".join(traceback.format_exception(*record.exc_info))
            res += self._format_tb(record.exc_text, "  ")
        if record.stack_info:
            res += self._format_tb(record.stack_info, "  ")

        res += yuio.term.Color.NONE

        return res

    def _format_tb(self, tb: str, indent: str) -> yuio.term.ColorizedString:
        highlighter = yuio.md.SyntaxHighlighter.get_highlighter("python-traceback")
        return highlighter.highlight(self.theme, tb).indent(indent, indent)

    def _format_task(self, task: Task) -> yuio.term.ColorizedString:
        res = yuio.term.ColorizedString()

        ctx = task._status.value

        if decoration := self.theme.msg_decorations.get("task"):
            res += self.theme.get_color(f"task/decoration:{ctx}")
            res += decoration

        res += self._format_task_msg(task)
        res += self.theme.get_color(f"task:{ctx}")
        res += " - "
        res += self.theme.get_color(f"task/progress:{ctx}")
        res += task._status.value
        res += self.theme.get_color(f"task:{ctx}")
        res += "\n"

        res += yuio.term.Color.NONE

        return res

    def _format_task_msg(self, task: Task) -> yuio.term.ColorizedString:
        if task._cached_msg is None:
            msg = self.formatter.colorize(
                task._msg, default_color=f"task/heading:{task._status.value}"
            )
            if task._args:
                msg %= task._args
            task._cached_msg = msg
        return task._cached_msg

    def _format_task_comment(self, task: Task) -> yuio.term.ColorizedString | None:
        if task._status is not Task._Status.RUNNING:
            return None
        if task._cached_comment is None and task._comment is not None:
            comment = self.formatter.colorize(
                task._comment, default_color=f"task/comment:{task._status.value}"
            )
            if task._comment_args:
                comment %= task._comment_args
            task._cached_comment = comment
        return task._cached_comment

    def _draw_task(self, task: Task, indent: int):
        self._tasks_printed += 1
        self._rc.move_pos(indent * 2, 0)
        self._draw_task_progressbar(task)
        self._rc.write(self._format_task_msg(task))
        self._draw_task_progress(task)
        if comment := self._format_task_comment(task):
            self._rc.set_color_path(f"task:{task._status.value}")
            self._rc.write(" - ")
            self._rc.write(comment)
        self._rc.new_line()

        for subtask in task._subtasks:
            self._draw_task(subtask, indent + 1)

    def _draw_task_progress(self, task: Task):
        if task._status in (Task._Status.DONE, Task._Status.ERROR):
            self._rc.set_color_path(f"task:{task._status.value}")
            self._rc.write(" - ")
            self._rc.set_color_path(f"task/progress:{task._status.value}")
            self._rc.write(task._status.name.lower())
        elif task._progress_done is not None:
            self._rc.set_color_path(f"task:{task._status.value}")
            self._rc.write(" - ")
            self._rc.set_color_path(f"task/progress:{task._status.value}")
            self._rc.write(task._progress_done)
            if task._progress_total is not None:
                self._rc.set_color_path(f"task:{task._status.value}")
                self._rc.write("/")
                self._rc.set_color_path(f"task/progress:{task._status.value}")
                self._rc.write(task._progress_total)

    def _draw_task_progressbar(self, task: Task):
        if task._status != Task._Status.RUNNING:
            self._rc.set_color_path(f"task/decoration:{task._status.value}")
            self._rc.write(self.theme.spinner_static_symbol)
        elif task._progress is None:
            self._rc.set_color_path(f"task/decoration:{task._status.value}")
            if self.theme.spinner_pattern:
                self._rc.write(
                    self.theme.spinner_pattern[
                        self._spinner_state % len(self.theme.spinner_pattern)
                    ]
                )
        else:
            total_width = self.theme.progress_bar_width
            done_width = round(max(0, min(1, task._progress)) * total_width)

            self._rc.set_color_path(f"task/progressbar:{task._status.value}")
            self._rc.write(self.theme.progress_bar_start_symbol)

            done_color = yuio.term.Color.lerp(
                self.theme.get_color("task/progressbar/done/start"),
                self.theme.get_color("task/progressbar/done/end"),
            )

            for i in range(0, done_width):
                self._rc.set_color(done_color((i + i / total_width) / total_width))
                self._rc.write(self.theme.progress_bar_done_symbol)

            pending_color = yuio.term.Color.lerp(
                self.theme.get_color("task/progressbar/pending/start"),
                self.theme.get_color("task/progressbar/pending/end"),
            )

            for i in range(done_width, total_width):
                self._rc.set_color(pending_color((i + i / total_width) / total_width))
                self._rc.write(self.theme.progress_bar_pending_symbol)

            self._rc.set_color_path(f"task/progressbar:{task._status.value}")
            self._rc.write(self.theme.progress_bar_end_symbol)

        self._rc.set_color_path(f"task:{task._status.value}")
        self._rc.write(" ")


class _WrappedOutput(_t.TextIO):
    def __init__(self, wrapped: _t.TextIO):
        self.__wrapped = wrapped

    @property
    def mode(self) -> str:
        return self.__wrapped.mode

    @property
    def name(self) -> str:
        return self.__wrapped.name

    def close(self):
        self.__wrapped.close()

    @property
    def closed(self) -> bool:
        return self.__wrapped.closed

    def fileno(self) -> int:
        return self.__wrapped.fileno()

    def flush(self):
        self.__wrapped.flush()

    def isatty(self) -> bool:
        return self.__wrapped.isatty()

    def writable(self) -> bool:
        return self.__wrapped.writable()

    def write(self, s: str, /) -> int:
        _manager().print_direct(s, self.__wrapped)
        return len(s)

    def writelines(self, lines: _t.Iterable[str], /):
        _manager().print_direct_lines(lines, self.__wrapped)

    def readable(self) -> bool:
        return self.__wrapped.readable()

    def read(self, n: int = -1) -> str:
        return self.__wrapped.read(n)

    def readline(self, limit: int = -1) -> str:
        return self.__wrapped.readline(limit)

    def readlines(self, hint: int = -1) -> list[str]:
        return self.__wrapped.readlines(hint)

    def seek(self, offset: int, whence: int = 0) -> int:
        return self.__wrapped.seek(offset, whence)

    def seekable(self) -> bool:
        return self.__wrapped.seekable()

    def tell(self) -> int:
        return self.__wrapped.tell()

    def truncate(self, size: int | None = None) -> int:
        return self.__wrapped.truncate(size)

    def __enter__(self) -> _t.TextIO:
        return self.__wrapped.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__wrapped.__exit__(exc_type, exc_val, exc_tb)

    @property
    def buffer(self) -> _t.BinaryIO:
        return self.__wrapped.buffer

    @property
    def encoding(self) -> str:
        return self.__wrapped.encoding

    @property
    def errors(self) -> str | None:
        return self.__wrapped.errors

    @property
    def line_buffering(self) -> int:
        return self.__wrapped.line_buffering

    @property
    def newlines(self) -> _t.Any:
        return self.__wrapped.newlines



# Yuio project, MIT license.
#
# https://github.com/taminomara/yuio/
#
# You're free to copy this file to your project and edit it for your needs,
# just keep this copyright line please :3

"""
This module provides basic functionality to interact with git.
It comes in handy when writing deployment scripts.


Interacting with a repository
-----------------------------

All repository interactions are done through the :class:`Repo` class
and its methods. If an interaction fails, a :class:`GitError` is raised.

.. autoclass:: Repo
    :members:

.. autoclass:: GitError


Status objects
--------------

:meth:`Repo.status` returns repository status parsed from ``git status`` command.
It can show changed and unmerged files and submodules. See details about change
representation in `git status`__ manual.

__ https://git-scm.com/docs/git-status#_output

Yuio represents ``git status`` output as close to the original as possible,
but makes some convenience renames. This results in somewhat unexpected
class structure:

.. raw:: html

    <p>
    <pre class="mermaid">
    ---
    config:
        class:
            hideEmptyMembersBox: true
    ---
    classDiagram

    class PathStatus
    click PathStatus href "#yuio.git.PathStatus" "yuio.git.PathStatus"

    class FileStatus
    click FileStatus href "#yuio.git.FileStatus" "yuio.git.FileStatus"
    PathStatus <|-- FileStatus

    class SubmoduleStatus
    click SubmoduleStatus href "#yuio.git.SubmoduleStatus" "yuio.git.SubmoduleStatus"
    FileStatus <|-- SubmoduleStatus

    class UnmergedFileStatus
    click UnmergedFileStatus href "#yuio.git.UnmergedFileStatus" "yuio.git.UnmergedFileStatus"
    PathStatus <|-- UnmergedFileStatus

    class UnmergedSubmoduleStatus
    click UnmergedSubmoduleStatus href "#yuio.git.UnmergedSubmoduleStatus" "yuio.git.UnmergedSubmoduleStatus"
    UnmergedFileStatus <|-- UnmergedSubmoduleStatus
    </pre>
    </p>

.. autoclass:: Status
    :members:

.. autoclass:: PathStatus
    :members:

.. autoclass:: FileStatus
    :members:

.. autoclass:: SubmoduleStatus
    :members:

.. autoclass:: UnmergedFileStatus
    :members:

.. autoclass:: UnmergedSubmoduleStatus
    :members:

.. autoclass:: Modification
    :members:


Commit objects
--------------

.. autoclass:: Commit
    :members:

.. autoclass:: CommitTrailers
    :members:


Parsing git refs
----------------

When you need to query a git ref from a user, :class:`RefParser` will check
that the given ref is formatted correctly. Use :class:`Ref` in your type hints
to help Yuio detect that you want to parse it as a git reference:

.. autoclass:: Ref

.. autoclass:: Tag

.. autoclass:: Branch

.. autoclass:: Remote

.. autoclass:: RefParser

.. autoclass:: TagParser

.. autoclass:: BranchParser

.. autoclass:: RemoteParser

If you know path to your repository before hand, and want to make sure that the user
supplies a valid ref that points to an existing git object, use :class:`CommitParser`:

.. autoclass:: CommitParser


Autocompleting git refs
-----------------------

.. autoclass:: RefCompleter

.. autoclass:: RefCompleterMode
    :members:

"""

from __future__ import annotations

import contextlib
import dataclasses
import enum
import functools
import logging
import pathlib
import re
import subprocess
import textwrap
from dataclasses import dataclass
from datetime import datetime

import yuio.complete
import yuio.parse
from yuio import _typing as _t

__all__ = [
    "GitError",
    "Ref",
    "Tag",
    "Branch",
    "Remote",
    "Repo",
    "Commit",
    "CommitTrailers",
    "Modification",
    "PathStatus",
    "FileStatus",
    "SubmoduleStatus",
    "UnmergedFileStatus",
    "UnmergedSubmoduleStatus",
    "Status",
    "RefCompleterMode",
    "RefCompleter",
    "RefParser",
    "TagParser",
    "BranchParser",
    "RemoteParser",
    "CommitParser",
]

_logger = logging.getLogger(__name__)


class GitError(subprocess.SubprocessError):
    """
    Raised when git returns a non-zero exit code.

    """


Ref = _t.NewType("Ref", str)
"""
A special kind of string that contains a git object reference.

Ref is not guaranteed to be valid; this type is used in type hints
to make use of the :class:`RefParser`.

"""

Tag = _t.NewType("Tag", str)
"""
A special kind of string that contains a tag name.

Ref is not guaranteed to be valid; this type is used in type hints
to make use of the :class:`TagParser`.

"""

Branch = _t.NewType("Branch", str)
"""
A special kind of string that contains a branch name.

Ref is not guaranteed to be valid; this type is used in type hints
to make use of the :class:`BranchParser`.

"""

Remote = _t.NewType("Remote", str)
"""
A special kind of string that contains a remote branch name.

Ref is not guaranteed to be valid; this type is used in type hints
to make use of the :class:`RemoteParser`.

"""


_LOG_FMT = "%H%n%aN%n%aE%n%aI%n%cN%n%cE%n%cI%n%(decorate:prefix=,suffix=,tag=,separator= )%n%w(0,0,1)%B%w(0,0)%n-"
_LOG_TRAILERS_FMT = "%H%n%w(0,1,1)%(trailers:only=true)%w(0,0)%n-"
_LOG_TRAILER_KEY_RE = re.compile(r"^(?P<key>\S+):\s")


class Repo:
    """
    A class that allows interactions with a git repository.

    :param path:
        path to the repo root dir.
    :param skip_checks:
        don't check if we're inside a repo.
    :param env:
        environment variables for the git executable.

    """

    def __init__(
        self,
        path: pathlib.Path | str,
        /,
        skip_checks: bool = False,
        env: dict[str, str] | None = None,
    ):
        self.__path = pathlib.Path(path)
        self.__env = env

        if skip_checks:
            return

        try:
            self.git("--version")
        except FileNotFoundError:
            raise GitError(f"git executable not found")

        try:
            self.git("rev-parse", "--is-inside-work-tree")
        except GitError:
            raise GitError(f"{path} is not a git repository")

    @property
    def path(self) -> pathlib.Path:
        """
        Path to the repo, as was passed to the constructor.

        """

        return self.__path

    def git(self, *args: str | pathlib.Path, capture_output: bool = True) -> bytes:
        """
        Call git and return its stdout.

        """

        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug("git %s", " ".join(map(str, args)))

        with contextlib.ExitStack() as context:
            if not capture_output:
                import yuio.io

                context.enter_context(yuio.io.SuspendOutput())
            try:
                res = subprocess.run(
                    ["git"] + list(args),
                    capture_output=capture_output,
                    cwd=self.__path,
                    env=self.__env,
                )
            except FileNotFoundError:
                raise GitError(f"git executable not found")

        if res.stderr:
            _logger.debug("%s", res.stderr.decode())

        if res.returncode != 0:
            raise GitError(
                f"git exited with status code {res.returncode}:\n"
                f"{res.stderr.decode()}"
            )

        return res.stdout

    @functools.cached_property
    def root(self) -> pathlib.Path:
        """
        The root directory of the repo.

        """

        return pathlib.Path(
            self.git("rev-parse", "--path-format=absolute", "--show-toplevel")
            .decode()
            .strip()
        ).resolve()

    @functools.cached_property
    def git_dir(self) -> pathlib.Path:
        """
        Get path to the ``.git`` directory of the repo.

        """

        return pathlib.Path(
            self.git("rev-parse", "--path-format=absolute", "--git-dir")
            .decode()
            .strip()
        ).resolve()

    def status(
        self, /, include_ignored: bool = False, include_submodules: bool = True
    ) -> Status:
        """
        Query the current repository status.

        :param include_ignored:
            include ignored status in the list of changes. Disable by default.

        :param include_submodules:
            include status of submodules in the list of changes. Enabled by default.

        """

        text = self.git(
            "status",
            "--porcelain=v2",
            "-z",
            "--ahead-behind",
            "--branch",
            "--renames",
            "--untracked-files=normal",
            "--ignore-submodules=" + ("none" if include_submodules else "all"),
            "--ignored=" + ("matching" if include_ignored else "no"),
        )
        lines = iter(text.split(b"\0"))

        status = Status(commit=None)

        for line_b in lines:
            line = line_b.decode()
            if line.startswith("# branch.oid"):
                if line[13:] != "(initial)":
                    status.commit = line[13:]
            elif line.startswith("# branch.head"):
                if line[14:] != "(detached)":
                    status.branch = line[14:]
            elif line.startswith("# branch.upstream"):
                status.upstream = line[18:]
            elif line.startswith("# branch.ab"):
                match = re.match(r"^\+(\d+) -(\d+)$", line[12:])
                assert match is not None
                status.ahead = int(match.group(1))
                status.behind = int(match.group(2))
            elif line.startswith("1"):
                match = re.match(
                    r"^(?P<X>.)(?P<Y>.) (?P<sub>.{4}) (?:[^ ]+ ){5}(?P<path>.*)$",
                    line[2:],
                )
                assert match is not None
                sub = match.group("sub")
                if sub[0] == "S":
                    path_status = SubmoduleStatus(
                        path=pathlib.Path(match.group("path")),
                        path_from=None,
                        staged=Modification(match.group("X")),
                        tree=Modification(match.group("Y")),
                        commit_changed=sub[1] != ".",
                        has_tracked_changes=sub[2] != ".",
                        has_untracked_changes=sub[3] != ".",
                    )
                else:
                    path_status = FileStatus(
                        path=pathlib.Path(match.group("path")),
                        path_from=None,
                        staged=Modification(match.group("X")),
                        tree=Modification(match.group("Y")),
                    )
                status.changes.append(path_status)
            elif line.startswith("2"):
                match = re.match(
                    r"^(?P<X>.)(?P<Y>.) (?P<sub>.{4}) (?:[^ ]+ ){6}(?P<path>.*)$",
                    line[2:],
                )
                assert match is not None
                path_from = pathlib.Path(next(lines).decode())
                sub = match.group("sub")
                if sub[0] == "S":
                    path_status = SubmoduleStatus(
                        path=pathlib.Path(match.group("path")),
                        path_from=path_from,
                        staged=Modification(match.group("X")),
                        tree=Modification(match.group("Y")),
                        commit_changed=sub[1] != ".",
                        has_tracked_changes=sub[2] != ".",
                        has_untracked_changes=sub[3] != ".",
                    )
                else:
                    path_status = FileStatus(
                        path=pathlib.Path(match.group("path")),
                        path_from=path_from,
                        staged=Modification(match.group("X")),
                        tree=Modification(match.group("Y")),
                    )
                status.changes.append(path_status)
            elif line.startswith("u"):
                match = re.match(
                    r"^(?P<X>.)(?P<Y>.) (?P<sub>.{4}) (?:[^ ]+ ){7}(?P<path>.*)$",
                    line[2:],
                )
                assert match is not None
                sub = match.group("sub")
                if sub[0] == "S":
                    path_status = UnmergedSubmoduleStatus(
                        path=pathlib.Path(match.group("path")),
                        us=Modification(match.group("X")),
                        them=Modification(match.group("Y")),
                        commit_changed=sub[1] != ".",
                        has_tracked_changes=sub[2] != ".",
                        has_untracked_changes=sub[3] != ".",
                    )
                else:
                    path_status = UnmergedFileStatus(
                        path=pathlib.Path(match.group("path")),
                        us=Modification(match.group("X")),
                        them=Modification(match.group("Y")),
                    )
                status.changes.append(path_status)
            elif line.startswith("?"):
                status.changes.append(
                    FileStatus(
                        path=pathlib.Path(line[2:]),
                        path_from=None,
                        staged=Modification.UNTRACKED,
                        tree=Modification.UNTRACKED,
                    )
                )
            elif line.startswith("!"):
                status.changes.append(
                    FileStatus(
                        path=pathlib.Path(line[2:]),
                        path_from=None,
                        staged=Modification.IGNORED,
                        tree=Modification.IGNORED,
                    )
                )

        try:
            status.cherry_pick_head = (
                self.git("rev-parse", "--verify", "CHERRY_PICK_HEAD").decode().strip()
            )
        except GitError:
            pass
        try:
            status.merge_head = (
                self.git("rev-parse", "--verify", "MERGE_HEAD").decode().strip()
            )
        except GitError:
            pass
        try:
            status.rebase_head = (
                self.git("rev-parse", "--verify", "REBASE_HEAD").decode().strip()
            )
        except GitError:
            pass
        try:
            status.revert_head = (
                self.git("rev-parse", "--verify", "REVERT_HEAD").decode().strip()
            )
        except GitError:
            pass

        return status

    def print_status(self):
        """
        Run ``git status`` and show its output to the user.

        """

        self.git("status", capture_output=False)

    def log(self, *refs: str, max_entries: int | None = None) -> list[Commit]:
        """
        Query the log for given git objects.

        :param refs:
            git references that will be passed to ``git log``.
        :param max_entries:
            maximum number of returned references.

        """

        args = [
            f"--pretty=format:{_LOG_FMT}",
            "--decorate-refs=refs/tags",
            "--decorate=short",
        ]

        if max_entries is not None:
            args += ["-n", str(max_entries)]

        args += list(refs)

        text = self.git("log", *args)
        lines = iter(text.decode().split("\n"))

        commits = []

        while commit := self.__parse_single_log_entry(lines):
            commits.append(commit)

        return commits

    def trailers(
        self, *refs: str, max_entries: int | None = None
    ) -> list[CommitTrailers]:
        """
        Query trailer lines for given git objects.

        Trailers are lines at the end of a commit message formatted as ``key: value``
        pairs. See `git-interpret-trailers`__ for further description.

        __ https://git-scm.com/docs/git-interpret-trailers

        :param refs:
            git references that will be passed to ``git log``.
        :param max_entries:
            maximum number of checked commits.

            .. warning::

                This option limits number of checked commits, not the number
                of trailers.

        """

        args = [f"--pretty=format:{_LOG_TRAILERS_FMT}"]

        if max_entries is not None:
            args += ["-n", str(max_entries)]

        args += list(refs)

        text = self.git("log", *args)
        lines = iter(text.decode().split("\n"))

        trailers = []

        while commit := self.__parse_single_trailer_entry(lines):
            trailers.append(commit)

        return trailers

    def show(self, ref: str, /) -> Commit | None:
        """
        Query information for the given git object.

        Return :data:`None` if object is not found.

        :param ref:
            git reference that will be passed to ``git log``.

        """

        try:
            self.git("rev-parse", "--verify", ref)
        except GitError:
            return None

        log = self.log(ref, max_entries=1)
        if not log:
            return None
        else:
            commit = log[0]
            commit.orig_ref = ref
            return commit

    @staticmethod
    def __parse_single_log_entry(lines) -> Commit | None:
        try:
            commit = next(lines)
            author = next(lines)
            author_email = next(lines)
            author_datetime = datetime.fromisoformat(next(lines).replace("Z", "+00:00"))
            committer = next(lines)
            committer_email = next(lines)
            committer_datetime = datetime.fromisoformat(
                next(lines).replace("Z", "+00:00")
            )
            tags = next(lines).split()
            title = next(lines)
            body = ""

            while True:
                line = next(lines)
                if not line or line.startswith(" "):
                    body += line[1:] + "\n"
                else:
                    break

            body = body.strip("\n")
            if body:
                body += "\n"

            return Commit(
                commit,
                tags,
                author,
                author_email,
                author_datetime,
                committer,
                committer_email,
                committer_datetime,
                title,
                body,
            )
        except StopIteration:
            return None

    @staticmethod
    def __parse_single_trailer_entry(
        lines,
    ) -> CommitTrailers | None:
        try:
            commit = next(lines)
            trailers = []
            current_key = None
            current_value = ""

            while True:
                line = next(lines)
                if not line or line.startswith(" "):
                    line = line[1:] + "\n"
                    if match := _LOG_TRAILER_KEY_RE.match(line):
                        if current_key:
                            first, *rest = current_value.splitlines(keepends=True)
                            current_value = (
                                first.strip() + "\n" + textwrap.dedent("".join(rest))
                            ).rstrip() + "\n"
                            trailers.append((current_key, current_value))
                        current_key = match.group("key")
                        current_value = line[match.end() :]
                    else:
                        current_value += line
                else:
                    break
            if current_key:
                first, *rest = current_value.splitlines(keepends=True)
                current_value = (
                    first.strip() + "\n" + textwrap.dedent("".join(rest))
                ).rstrip() + "\n"
                trailers.append((current_key, current_value))

            return CommitTrailers(commit, trailers)
        except StopIteration:
            return None

    def tags(self) -> list[str]:
        """
        List all tags in this repository.

        """

        return (
            self.git("for-each-ref", "--format=%(refname:short)", "refs/tags")
            .decode()
            .splitlines()
        )

    def branches(self) -> list[str]:
        """
        List all branches in this repository.

        """

        return (
            self.git("for-each-ref", "--format=%(refname:short)", "refs/heads")
            .decode()
            .splitlines()
        )

    def remotes(self) -> list[str]:
        """
        List all remote branches in this repository.

        """

        return [
            remote
            for remote in self.git(
                "for-each-ref", "--format=%(refname:short)", "refs/remotes"
            )
            .decode()
            .splitlines()
            if "/" in remote
        ]


@dataclass
class Commit:
    """
    Commit description.

    """

    hash: str
    """
    Commit hash.

    """

    tags: list[str]
    """
    Tags attached to this commit.

    """

    author: str
    """
    Author name.

    """

    author_email: str
    """
    Author email.

    """

    author_datetime: datetime
    """
    Author time.

    """

    committer: str
    """
    Committer name.

    """

    committer_email: str
    """
    Committer email.

    """

    committer_datetime: datetime
    """
    Committer time.

    """

    title: str
    """
    Commit title, i.e. first line of the message.

    """

    body: str
    """
    Commit body, i.e. the rest of the message.

    """

    orig_ref: str | None = None
    """
    If commit was parsed from a user input, this field will contain
    original input. I.e. if a user enters ``HEAD`` and it gets resolved
    into a commit, `orig_ref` will contain string ``"HEAD"``.

    See also :class:`CommitParser`.

    """

    @property
    def short_hash(self):
        """
        First seven characters of the commit hash.

        """

        return self.hash[:7]

    def __str__(self):
        if self.orig_ref:
            return self.orig_ref
        else:
            return self.short_hash


@dataclass
class CommitTrailers:
    """
    Commit trailers.

    """

    hash: str
    """
    Commit hash.

    """

    trailers: list[tuple[str, str]]
    """
    Key-value pairs for commit trailers.

    """


class Modification(enum.Enum):
    """
    For changed file or submodule, what modification was applied to it.

    """

    UNMODIFIED = "."
    """
    File wasn't changed.

    """

    MODIFIED = "M"
    """
    File was changed.

    """

    SUBMODULE_MODIFIED = "m"
    """
    Contents of submodule were modified.

    """

    TYPE_CHANGED = "T"
    """
    File type changed.

    """

    ADDED = "A"
    """
    File was created.

    """

    DELETED = "D"
    """
    File was deleted.

    """

    RENAMED = "R"
    """
    File was renamed (and possibly changed).

    """

    COPIED = "C"
    """
    File was copied (and possibly changed).

    """

    UPDATED = "U"
    """
    File was updated but unmerged.

    """

    UNTRACKED = "?"
    """
    File is untracked, i.e. not yet staged or committed.

    """

    IGNORED = "!"
    """
    File is in ``.gitignore``.

    """


@dataclass
class PathStatus:
    """
    Status of a changed path.

    """

    path: pathlib.Path
    """
    Path of the file.

    """


@dataclass
class FileStatus(PathStatus):
    """
    Status of a changed file.

    """

    path_from: pathlib.Path | None
    """
    If file was moved, contains path where it was moved from.

    """

    staged: Modification
    """
    File modification in the index (staged).

    """

    tree: Modification
    """
    File modification in the tree (unstaged).

    """


@dataclass
class SubmoduleStatus(FileStatus):
    """
    Status of a submodule.

    """

    commit_changed: bool
    """
    The submodule has a different HEAD than recorded in the index.

    """

    has_tracked_changes: bool
    """
    Tracked files were changed in the submodule.

    """

    has_untracked_changes: bool
    """
    Untracked files were changed in the submodule.

    """


@dataclass
class UnmergedFileStatus(PathStatus):
    """
    Status of an unmerged file.

    """

    us: Modification
    """
    File modification that has happened at the head.

    """

    them: Modification
    """
    File modification that has happened at the merge head.

    """


@dataclass
class UnmergedSubmoduleStatus(UnmergedFileStatus):
    """
    Status of an unmerged submodule.

    """

    commit_changed: bool
    """
    The submodule has a different HEAD than recorded in the index.

    """

    has_tracked_changes: bool
    """
    Tracked files were changed in the submodule.

    """

    has_untracked_changes: bool
    """
    Untracked files were changed in the submodule.

    """


@dataclass
class Status:
    """
    Status of a working copy.

    """

    commit: str | None
    """
    Current commit hash. Can be absent if current branch is orphaned and doesn't have
    any commits yet.

    """

    branch: str | None = None
    """
    Name of the current branch.

    """

    upstream: str | None = None
    """
    Name of the upstream branch.

    """

    ahead: int | None = None
    """
    Number of commits the branch is ahead of upstream.

    """

    behind: int | None = None
    """
    Number of commits the branch is behind of upstream.

    """

    changes: list[PathStatus] = dataclasses.field(default_factory=list)
    """
    List of changed files, both tracked and untracked.

    See details about change representation in `git status`__ manual.

    __ https://git-scm.com/docs/git-status#_output

    """

    cherry_pick_head: str | None = None
    """
    Position of the ``CHERRY_PICK_HEAD``.

    If this field is not :data:`None`, cherry pick is in progress.

    """

    merge_head: str | None = None
    """
    Position of the ``MERGE_HEAD``.

    If this field is not :data:`None`, merge is in progress.

    """

    rebase_head: str | None = None
    """
    Position of the ``REBASE_HEAD``.

    If this field is not :data:`None`, rebase is in progress.

    """

    revert_head: str | None = None
    """
    Position of the ``REVERT_HEAD``.

    If this field is not :data:`None`, revert is in progress.

    """

    def has_staged_changes(self) -> bool:
        """
        Return :data:`True` if there are unstaged changes in this repository.

        """

        return next(self.get_staged_changes(), None) is not None

    def get_staged_changes(self) -> _t.Iterator[PathStatus]:
        return (
            change
            for change in self.changes
            if isinstance(change, FileStatus)
            and change.staged
            not in [
                Modification.UNMODIFIED,
                Modification.IGNORED,
                Modification.UNTRACKED,
            ]
        )

    def has_unstaged_changes(self) -> bool:
        """
        Return :data:`True` if there are unstaged changes in this repository.

        """

        return next(self.get_unstaged_changes(), None) is not None

    def get_unstaged_changes(self) -> _t.Iterator[PathStatus]:
        return (
            change
            for change in self.changes
            if isinstance(change, UnmergedFileStatus)
            or (
                isinstance(change, FileStatus)
                and change.tree
                not in [
                    Modification.UNMODIFIED,
                    Modification.IGNORED,
                ]
            )
        )


class RefCompleterMode(enum.Enum):
    """
    Specifies operation modes for :class:`RefCompleter`.

    """

    BRANCH = "b"
    """
    Completes branches.

    """

    REMOTE = "r"
    """
    Completes remote branches.

    """

    TAG = "t"
    """
    Completes tags.

    """

    HEAD = "h"
    """
    Completes ``HEAD`` and ``ORIG_HEAD``.

    """


class RefCompleter(yuio.complete.Completer):
    """
    Completes git refs.

    :param repo:
        source of completions.
    :param modes:
        which objects to complete.

    """

    def __init__(self, repo: Repo, modes: set[RefCompleterMode] | None = None):
        super().__init__()

        self._repo = repo
        self._modes = modes or {
            RefCompleterMode.BRANCH,
            RefCompleterMode.TAG,
            RefCompleterMode.HEAD,
        }

    def _process(self, collector: yuio.complete.CompletionCollector, /):
        try:
            if RefCompleterMode.HEAD in self._modes:
                collector.add_group()
                git_dir = self._repo.git_dir
                for head in ["HEAD", "ORIG_HEAD"]:
                    if (git_dir / head).exists():
                        collector.add(head)
            if RefCompleterMode.BRANCH in self._modes:
                collector.add_group()
                for branch in self._repo.branches():
                    collector.add(branch, comment="branch")
            if RefCompleterMode.REMOTE in self._modes:
                collector.add_group()
                for remote in self._repo.remotes():
                    collector.add(remote, comment="remote")
            if RefCompleterMode.TAG in self._modes:
                collector.add_group()
                for tag in self._repo.tags():
                    collector.add(tag, comment="tag")
        except GitError:
            pass

    def _get_completion_model(
        self, *, is_many: bool = False
    ) -> yuio.complete._CompleterSerializer.Model:
        return yuio.complete._CompleterSerializer.Git(
            {
                yuio.complete._CompleterSerializer.Git.Mode(mode.value)
                for mode in self._modes
            }
        )


def CommitParser(*, repo: Repo) -> yuio.parse.Parser[Commit]:
    """
    A parser for git refs (commits, tags, branches, and so on).

    This parser validates that the given ref exists in the given repository,
    parses it and returns a commit data associated with this ref.

    If you need a simple string without additional validation,
    use :class:`RefParser`.

    :param repo:
        initialized repository is required to ensure that commit is valid.

    """

    def map(value: str, /) -> Commit:
        commit = repo.show(value)
        if commit is None:
            raise yuio.parse.ParsingError("invalid git ref")
        return commit

    def rev(value: Commit | object) -> str:
        if isinstance(value, Commit):
            return str(value)
        else:
            raise TypeError(
                f"parser Commit can't handle value "
                f"of type {_t.type_repr(type(value))}"
            )

    return yuio.parse.WithMeta(
        yuio.parse.Map(yuio.parse.Str(), map, rev),
        desc="<commit>",
        completer=RefCompleter(repo),
    )


T = _t.TypeVar("T")


class _RefParserImpl(yuio.parse.Str, _t.Generic[T]):
    @functools.cached_property
    def _description(self):
        return "<" + self.__class__.__name__.removesuffix("Parser").lower() + ">"

    def describe(self) -> str | None:
        return self._description

    def describe_or_def(self) -> str:
        return self._description

    def describe_many(self) -> str | tuple[str, ...] | None:
        return self._description

    def describe_many_or_def(self) -> str | tuple[str, ...]:
        return self._description


class RefParser(_RefParserImpl[Ref]):
    """
    A parser that provides autocompletion for git refs, but doesn't verify
    anything else.

    """

    def completer(self) -> yuio.complete.Completer:
        return RefCompleter(
            Repo(pathlib.Path.cwd(), skip_checks=True),
        )


class TagParser(_RefParserImpl[Tag]):
    """
    A parser that provides autocompletion for git tag, but doesn't verify
    anything else.

    """

    def completer(self) -> yuio.complete.Completer:
        return RefCompleter(
            Repo(pathlib.Path.cwd(), skip_checks=True),
            {RefCompleterMode.TAG},
        )


class BranchParser(_RefParserImpl[Branch]):
    """
    A parser that provides autocompletion for git branches, but doesn't verify
    anything else.

    """

    def completer(self) -> yuio.complete.Completer:
        return RefCompleter(
            Repo(pathlib.Path.cwd(), skip_checks=True),
            {RefCompleterMode.BRANCH},
        )


class RemoteParser(_RefParserImpl[Remote]):
    """
    A parser that provides autocompletion for git remotes, but doesn't verify
    anything else.

    """

    def completer(self) -> yuio.complete.Completer:
        return RefCompleter(
            Repo(pathlib.Path.cwd(), skip_checks=True),
            {RefCompleterMode.REMOTE},
        )


yuio.parse.register_type_hint_conversion(
    lambda ty, origin, args: RefParser() if ty is Ref else None
)

yuio.parse.register_type_hint_conversion(
    lambda ty, origin, args: TagParser() if ty is Tag else None
)

yuio.parse.register_type_hint_conversion(
    lambda ty, origin, args: BranchParser() if ty is Branch else None
)

yuio.parse.register_type_hint_conversion(
    lambda ty, origin, args: RemoteParser() if ty is Remote else None
)



import datetime
import os
import pathlib
import subprocess
import tempfile

import pytest

import yuio.complete
import yuio.git
import yuio.json_schema


class EqualsToAnything:
    def __eq__(self, value: object) -> bool:
        return True

    def __ne__(self, value: object) -> bool:
        return False

    def __lt__(self, rhs: object):
        return False


@pytest.fixture(autouse=True)
def git_env(monkeypatch):
    monkeypatch.setenv("GIT_CONFIG_GLOBAL", "")
    monkeypatch.setenv("GIT_AUTHOR_NAME", "author")
    monkeypatch.setenv("GIT_AUTHOR_EMAIL", "author@example.com")
    monkeypatch.setenv("GIT_AUTHOR_DATE", "2020-01-01T00:00:00Z")
    monkeypatch.setenv("GIT_COMMITTER_NAME", "committer")
    monkeypatch.setenv("GIT_COMMITTER_EMAIL", "committer@example.com")
    monkeypatch.setenv("GIT_COMMITTER_DATE", "2020-01-02T00:00:00Z")
    pass


@pytest.fixture()
def repo_base():
    with tempfile.TemporaryDirectory() as base_s:
        base = pathlib.Path(base_s).resolve()

        template = base / "template"
        template.mkdir()

        yield base


@pytest.fixture()
def repo_path(repo_base):
    repo = repo_base / "repo"
    repo.mkdir()

    subprocess.check_call(
        ["git", "init", ".", "-b", "main", "--template", str(repo_base / "template")],
        cwd=repo,
    )

    yield repo


@pytest.fixture()
def remote_repo_path(repo_base):
    repo = repo_base / "repo_remote"
    repo.mkdir()

    subprocess.check_call(
        ["git", "init", ".", "-b", "main", "--template", str(repo_base / "template")],
        cwd=repo,
    )
    subprocess.check_call(
        ["git", "commit", "--allow-empty", "--message", "initial commit"],
        cwd=repo,
    )

    yield repo


@pytest.fixture()
def repo(repo_path):
    return yuio.git.Repo(repo_path)


def test_not_a_repo():
    with tempfile.TemporaryDirectory() as base:
        with pytest.raises(yuio.git.GitError, match="not a git repository"):
            yuio.git.Repo(base)


def test_not_a_repo_skip_checks():
    with tempfile.TemporaryDirectory() as base:
        repo = yuio.git.Repo(base, skip_checks=True)
        with pytest.raises(yuio.git.GitError):
            repo.status()


@pytest.mark.skipif(os.name == "nt", reason="windows")
def test_git_unavailable(repo_path):
    with pytest.raises(yuio.git.GitError, match="git executable not found"):
        yuio.git.Repo(repo_path, env={"PATH": ""})


@pytest.mark.skipif(os.name == "nt", reason="windows")
def test_git_unavailable_skip_checks(repo_path):
    repo = yuio.git.Repo(repo_path, env={"PATH": ""}, skip_checks=True)
    with pytest.raises(yuio.git.GitError, match="git executable not found"):
        repo.status()


def test_simple(repo_path):
    repo = yuio.git.Repo(repo_path)
    assert repo.path == repo_path
    assert repo.root == repo_path
    assert repo.git_dir == repo_path / ".git"


def test_status(repo):
    status = repo.status()

    assert status.commit is None
    assert status.branch == "main"
    assert status.upstream is None
    assert status.ahead is None
    assert status.behind is None
    assert not status.has_staged_changes()
    assert not status.has_unstaged_changes()
    assert status.cherry_pick_head is None
    assert status.merge_head is None
    assert status.rebase_head is None
    assert status.revert_head is None
    assert status.changes == []

    repo.root.joinpath("foo").write_text("a")
    repo.root.joinpath("bar").write_text("b")

    status = repo.status()
    assert not status.has_staged_changes()
    assert status.has_unstaged_changes()
    assert status.changes == [
        yuio.git.FileStatus(
            path=pathlib.Path("bar"),
            path_from=None,
            staged=yuio.git.Modification.UNTRACKED,
            tree=yuio.git.Modification.UNTRACKED,
        ),
        yuio.git.FileStatus(
            path=pathlib.Path("foo"),
            path_from=None,
            staged=yuio.git.Modification.UNTRACKED,
            tree=yuio.git.Modification.UNTRACKED,
        ),
    ]

    repo.git("add", repo.root / "bar")

    status = repo.status()
    assert status.has_staged_changes()
    assert status.has_unstaged_changes()
    assert status.changes == [
        yuio.git.FileStatus(
            path=pathlib.Path("bar"),
            path_from=None,
            staged=yuio.git.Modification.ADDED,
            tree=yuio.git.Modification.UNMODIFIED,
        ),
        yuio.git.FileStatus(
            path=pathlib.Path("foo"),
            path_from=None,
            staged=yuio.git.Modification.UNTRACKED,
            tree=yuio.git.Modification.UNTRACKED,
        ),
    ]

    repo.git("add", repo.root / "foo")

    status = repo.status()
    assert status.has_staged_changes()
    assert not status.has_unstaged_changes()
    assert status.changes == [
        yuio.git.FileStatus(
            path=pathlib.Path("bar"),
            path_from=None,
            staged=yuio.git.Modification.ADDED,
            tree=yuio.git.Modification.UNMODIFIED,
        ),
        yuio.git.FileStatus(
            path=pathlib.Path("foo"),
            path_from=None,
            staged=yuio.git.Modification.ADDED,
            tree=yuio.git.Modification.UNMODIFIED,
        ),
    ]

    repo.git("commit", "-m", "message")

    status = repo.status()
    assert status.commit is not None
    assert not status.has_staged_changes()
    assert not status.has_unstaged_changes()
    assert status.changes == []


def test_status_ignored_files(repo):
    repo.root.joinpath(".gitignore").write_text("ignored.txt\n")
    repo.git("add", ".gitignore")
    repo.git("commit", "-m", "message")

    status = repo.status()

    assert status.branch == "main"
    assert status.upstream is None
    assert status.ahead is None
    assert status.behind is None
    assert not status.has_staged_changes()
    assert not status.has_unstaged_changes()
    assert status.cherry_pick_head is None
    assert status.merge_head is None
    assert status.rebase_head is None
    assert status.revert_head is None
    assert status.changes == []

    repo.root.joinpath("ignored.txt").write_text("...")

    status = repo.status()
    assert not status.has_staged_changes()
    assert not status.has_unstaged_changes()
    assert status.changes == []

    status = repo.status(include_ignored=True)
    assert not status.has_staged_changes()
    assert not status.has_unstaged_changes()
    assert status.changes == [
        yuio.git.FileStatus(
            path=pathlib.Path("ignored.txt"),
            path_from=None,
            staged=yuio.git.Modification.IGNORED,
            tree=yuio.git.Modification.IGNORED,
        ),
    ]


def test_status_remote(repo, remote_repo_path):
    repo.git("remote", "add", "origin", str(remote_repo_path))
    repo.git("fetch", "origin")
    repo.git("checkout", "main")

    status = repo.status()
    assert status.branch == "main"
    assert status.upstream == "origin/main"
    assert status.ahead == 0
    assert status.behind == 0

    repo.git("commit", "-m", "message", "--allow-empty")

    status = repo.status()
    assert status.branch == "main"
    assert status.upstream == "origin/main"
    assert status.ahead == 1
    assert status.behind == 0


def test_status_conflicts(repo):
    repo.git("commit", "-m", "message", "--allow-empty")
    repo.root.joinpath("file.txt").write_text("1")
    repo.git("add", "file.txt")
    repo.git("commit", "-m", "message")
    repo.git("checkout", "HEAD~1")
    repo.root.joinpath("file.txt").write_text("2")
    repo.git("add", "file.txt")
    repo.git("commit", "-m", "message")
    try:
        repo.git("merge", "main")
    except yuio.git.GitError:
        pass
    status = repo.status()
    assert status.commit is not None
    assert status.branch is None
    assert status.upstream is None
    assert status.ahead is None
    assert status.behind is None
    assert not status.has_staged_changes()
    assert status.has_unstaged_changes()
    assert status.cherry_pick_head is None
    assert status.merge_head is not None
    assert status.rebase_head is None
    assert status.revert_head is None
    assert status.changes == [
        yuio.git.UnmergedFileStatus(
            path=pathlib.Path("file.txt"),
            us=yuio.git.Modification.ADDED,
            them=yuio.git.Modification.ADDED,
        )
    ]


def test_status_move(repo):
    repo.root.joinpath("file.txt").write_text("foo bar baz" * 10)
    repo.git("add", "file.txt")
    repo.git("commit", "-m", "message", "--allow-empty")
    repo.git("mv", "file.txt", "file2.txt")
    status = repo.status()
    assert status.has_staged_changes()
    assert not status.has_unstaged_changes()
    assert status.changes == [
        yuio.git.FileStatus(
            path=pathlib.Path("file2.txt"),
            path_from=pathlib.Path("file.txt"),
            staged=yuio.git.Modification.RENAMED,
            tree=yuio.git.Modification.UNMODIFIED,
        )
    ]


def test_status_submodule(repo, remote_repo_path):
    repo.git(
        "-c",
        "protocol.file.allow=always",
        "submodule",
        "add",
        remote_repo_path,
        "submodule",
    )
    status = repo.status()
    assert status.commit is None
    assert status.branch == "main"
    assert status.upstream is None
    assert status.ahead is None
    assert status.behind is None
    assert status.has_staged_changes()
    assert not status.has_unstaged_changes()
    assert status.cherry_pick_head is None
    assert status.merge_head is None
    assert status.rebase_head is None
    assert status.revert_head is None
    assert status.changes == [
        yuio.git.FileStatus(
            path=pathlib.Path(".gitmodules"),
            path_from=None,
            staged=yuio.git.Modification.ADDED,
            tree=yuio.git.Modification.UNMODIFIED,
        ),
        yuio.git.SubmoduleStatus(
            path=pathlib.Path("submodule"),
            path_from=None,
            staged=yuio.git.Modification.ADDED,
            tree=yuio.git.Modification.UNMODIFIED,
            commit_changed=False,
            has_tracked_changes=False,
            has_untracked_changes=False,
        ),
    ]


def test_status_submodule_changed(repo, remote_repo_path):
    repo.git(
        "-c",
        "protocol.file.allow=always",
        "submodule",
        "add",
        remote_repo_path,
        "submodule",
    )
    repo.git("commit", "--message", "message")
    repo.root.joinpath("submodule/file.txt").write_text("foo")

    status = repo.status()
    assert status.commit is not None
    assert status.branch == "main"
    assert status.upstream is None
    assert status.ahead is None
    assert status.behind is None
    assert not status.has_staged_changes()
    assert status.has_unstaged_changes()
    assert status.cherry_pick_head is None
    assert status.merge_head is None
    assert status.rebase_head is None
    assert status.revert_head is None
    assert status.changes == [
        yuio.git.SubmoduleStatus(
            path=pathlib.Path("submodule"),
            path_from=None,
            staged=yuio.git.Modification.UNMODIFIED,
            tree=yuio.git.Modification.MODIFIED,
            commit_changed=False,
            has_tracked_changes=False,
            has_untracked_changes=True,
        )
    ]


def test_status_submodule_moved(repo, remote_repo_path):
    repo.git(
        "-c",
        "protocol.file.allow=always",
        "submodule",
        "add",
        remote_repo_path,
        "submodule",
    )
    repo.git("commit", "--message", "message")
    repo.git("mv", "submodule", "submodule_moved")

    status = repo.status()
    assert status.commit is not None
    assert status.branch == "main"
    assert status.upstream is None
    assert status.ahead is None
    assert status.behind is None
    assert status.has_staged_changes()
    assert not status.has_unstaged_changes()
    assert status.cherry_pick_head is None
    assert status.merge_head is None
    assert status.rebase_head is None
    assert status.revert_head is None
    assert status.changes == [
        yuio.git.FileStatus(
            path=pathlib.Path(".gitmodules"),
            path_from=None,
            staged=yuio.git.Modification.MODIFIED,
            tree=yuio.git.Modification.UNMODIFIED,
        ),
        yuio.git.SubmoduleStatus(
            path=pathlib.Path("submodule_moved"),
            path_from=pathlib.Path("submodule"),
            staged=yuio.git.Modification.RENAMED,
            tree=yuio.git.Modification.UNMODIFIED,
            commit_changed=False,
            has_tracked_changes=False,
            has_untracked_changes=False,
        ),
    ]


def test_status_submodule_conflicts(repo, remote_repo_path):
    repo.git(
        "-c",
        "protocol.file.allow=always",
        "submodule",
        "add",
        remote_repo_path,
        "submodule",
    )
    repo.git("commit", "--message", "message")

    submodule_repo = yuio.git.Repo(repo.root / "submodule")

    submodule_repo.root.joinpath("foo.txt").write_text("foo")
    submodule_repo.git("add", "foo.txt")
    submodule_repo.git("commit", "--message", "message")
    repo.git("add", "submodule")
    repo.git("commit", "--message", "message")

    repo.git("checkout", "HEAD~1")
    submodule_repo.git("checkout", "HEAD~1")

    submodule_repo.root.joinpath("foo.txt").write_text("bar")
    submodule_repo.git("add", "foo.txt")
    submodule_repo.git("commit", "--message", "message")
    repo.git("add", "submodule")
    repo.git("commit", "--message", "message")

    try:
        repo.git("merge", "main")
    except yuio.git.GitError:
        pass

    status = repo.status()
    assert status.commit is not None
    assert status.branch is None
    assert status.upstream is None
    assert status.ahead is None
    assert status.behind is None
    assert not status.has_staged_changes()
    assert status.has_unstaged_changes()
    assert status.cherry_pick_head is None
    assert status.merge_head is not None
    assert status.rebase_head is None
    assert status.revert_head is None
    assert status.changes == [
        yuio.git.UnmergedSubmoduleStatus(
            path=pathlib.Path("submodule"),
            us=yuio.git.Modification.UPDATED,
            them=yuio.git.Modification.UPDATED,
            commit_changed=False,
            has_tracked_changes=False,
            has_untracked_changes=False,
        )
    ]


def test_log(repo):
    repo.git(
        "commit",
        "--allow-empty",
        "--message",
        "Title 1\n\nBody 1",
    )
    repo.git(
        "commit",
        "--allow-empty",
        "--message",
        "Title 2\n\nBody 2",
    )
    repo.git("tag", "tag1")
    repo.git("tag", "tag2")
    repo.git(
        "commit",
        "--allow-empty",
        "--message",
        "Title 3\n\nBody 3",
        "--trailer",
        "Trailer1:Value1",
        "--trailer",
        "Trailer2:Value2",
    )

    log = repo.log()
    assert log == [
        yuio.git.Commit(
            hash=EqualsToAnything(),  # type: ignore
            tags=[],
            author="author",
            author_email="author@example.com",
            author_datetime=datetime.datetime(
                2020, 1, 1, 0, 0, tzinfo=datetime.timezone.utc
            ),
            committer="committer",
            committer_email="committer@example.com",
            committer_datetime=datetime.datetime(
                2020, 1, 2, 0, 0, tzinfo=datetime.timezone.utc
            ),
            title="Title 3",
            body="Body 3\n\nTrailer1: Value1\nTrailer2: Value2\n",
            orig_ref=None,
        ),
        yuio.git.Commit(
            hash=EqualsToAnything(),  # type: ignore
            tags=["tag2", "tag1"],
            author="author",
            author_email="author@example.com",
            author_datetime=datetime.datetime(
                2020, 1, 1, 0, 0, tzinfo=datetime.timezone.utc
            ),
            committer="committer",
            committer_email="committer@example.com",
            committer_datetime=datetime.datetime(
                2020, 1, 2, 0, 0, tzinfo=datetime.timezone.utc
            ),
            title="Title 2",
            body="Body 2\n",
            orig_ref=None,
        ),
        yuio.git.Commit(
            hash=EqualsToAnything(),  # type: ignore
            tags=[],
            author="author",
            author_email="author@example.com",
            author_datetime=datetime.datetime(
                2020, 1, 1, 0, 0, tzinfo=datetime.timezone.utc
            ),
            committer="committer",
            committer_email="committer@example.com",
            committer_datetime=datetime.datetime(
                2020, 1, 2, 0, 0, tzinfo=datetime.timezone.utc
            ),
            title="Title 1",
            body="Body 1\n",
            orig_ref=None,
        ),
    ]

    trailers = repo.trailers()
    assert trailers == [
        yuio.git.CommitTrailers(
            hash=log[0].hash,
            trailers=[("Trailer1", "Value1\n"), ("Trailer2", "Value2\n")],
        ),
        yuio.git.CommitTrailers(hash=log[1].hash, trailers=[]),
        yuio.git.CommitTrailers(hash=log[2].hash, trailers=[]),
    ]


def test_show(repo):
    repo.git("commit", "--allow-empty", "--message", "message")
    commit = repo.show("HEAD")
    assert commit == yuio.git.Commit(
        hash=EqualsToAnything(),  # type: ignore
        tags=[],
        author="author",
        author_email="author@example.com",
        author_datetime=datetime.datetime(
            2020, 1, 1, 0, 0, tzinfo=datetime.timezone.utc
        ),
        committer="committer",
        committer_email="committer@example.com",
        committer_datetime=datetime.datetime(
            2020, 1, 2, 0, 0, tzinfo=datetime.timezone.utc
        ),
        title="message",
        body="",
        orig_ref="HEAD",
    )

    assert repo.show("WAT") is None


def test_tags(repo):
    repo.git("commit", "--allow-empty", "--message", "message")
    repo.git("tag", "tag1")
    repo.git("tag", "tag2")
    assert set(repo.tags()) == {"tag1", "tag2"}


def test_branches(repo):
    repo.git("commit", "--allow-empty", "--message", "message")
    assert repo.branches() == ["main"]
    repo.git("checkout", "-B", "branch")
    assert set(repo.branches()) == {"branch", "main"}


def test_remotes(repo, remote_repo_path):
    repo.git("commit", "--allow-empty", "--message", "message")
    assert repo.remotes() == []
    repo.git("remote", "add", "origin", str(remote_repo_path))
    repo.git("fetch", "origin")
    assert repo.remotes() == ["origin/main"]


class TestRefCompleter:
    @pytest.fixture(autouse=True)
    def setup_repo(self, repo, remote_repo_path):
        repo.git("commit", "--allow-empty", "--message", "message")
        repo.git("tag", "tag1")
        repo.git("tag", "tag2")
        repo.git("remote", "add", "origin", str(remote_repo_path))
        repo.git("fetch", "origin")

    def test_default_mode(self, repo):
        completer = yuio.git.RefCompleter(repo)
        result = completer.complete("", 0)
        assert result == [
            yuio.complete.Completion(
                iprefix="",
                completion="HEAD",
                rsuffix="",
                rsymbols="",
                isuffix="",
                comment=None,
                dprefix="",
                dsuffix="",
                group_color_tag=None,
                group_id=EqualsToAnything(),
            ),
            yuio.complete.Completion(
                iprefix="",
                completion="main",
                rsuffix="",
                rsymbols="",
                isuffix="",
                comment="branch",
                dprefix="",
                dsuffix="",
                group_color_tag=None,
                group_id=EqualsToAnything(),
            ),
            yuio.complete.Completion(
                iprefix="",
                completion="tag1",
                rsuffix="",
                rsymbols="",
                isuffix="",
                comment="tag",
                dprefix="",
                dsuffix="",
                group_color_tag=None,
                group_id=EqualsToAnything(),
            ),
            yuio.complete.Completion(
                iprefix="",
                completion="tag2",
                rsuffix="",
                rsymbols="",
                isuffix="",
                comment="tag",
                dprefix="",
                dsuffix="",
                group_color_tag=None,
                group_id=EqualsToAnything(),
            ),
        ]

    def test_mode(self, repo):
        completer = yuio.git.RefCompleter(
            repo,
            modes={yuio.git.RefCompleterMode.HEAD, yuio.git.RefCompleterMode.REMOTE},
        )
        result = completer.complete("", 0)
        assert result == [
            yuio.complete.Completion(
                iprefix="",
                completion="HEAD",
                rsuffix="",
                rsymbols="",
                isuffix="",
                comment=None,
                dprefix="",
                dsuffix="",
                group_color_tag=None,
                group_id=EqualsToAnything(),
            ),
            yuio.complete.Completion(
                iprefix="",
                completion="origin/main",
                rsuffix="",
                rsymbols="",
                isuffix="",
                comment="remote",
                dprefix="",
                dsuffix="",
                group_color_tag=None,
                group_id=EqualsToAnything(),
            ),
        ]


class TestCommitParser:
    @pytest.fixture(autouse=True)
    def setup_repo(self, repo):
        repo.git("commit", "--allow-empty", "--message", "message")

    def test_basics(self, repo):
        parser = yuio.git.CommitParser(repo=repo)
        assert not parser.supports_parse_many()
        assert parser.get_nargs() == None
        with pytest.raises(RuntimeError):
            assert parser.parse_many(["HEAD"])
        assert parser.describe() == "<commit>"
        assert parser.describe_or_def() == "<commit>"
        assert parser.describe_many() == "<commit>"
        assert parser.describe_many_or_def() == "<commit>"
        assert parser.describe_value(repo.show("HEAD")) is None
        assert parser.describe_value_or_def(repo.show("HEAD")) == "HEAD"

    def test_json_schema(self, repo):
        parser = yuio.git.CommitParser(repo=repo)
        assert (
            parser.to_json_schema(yuio.json_schema.JsonSchemaContext())
            == yuio.json_schema.String()
        )
        assert parser.to_json_value(repo.show("HEAD")) == "HEAD"

    def test_parse(self, repo):
        parser = yuio.git.CommitParser(repo=repo)
        assert parser.parse("HEAD") == repo.show("HEAD")
        assert parser.parse_config("HEAD") == repo.show("HEAD")
        with pytest.raises(ValueError, match="invalid git ref"):
            parser.parse("WAT")

    def test_completer(self, repo):
        parser = yuio.git.CommitParser(repo=repo)
        assert isinstance(parser.completer(), yuio.git.RefCompleter)


class TestRefParser:
    @pytest.mark.parametrize(
        "parser,name,modes",
        [
            (
                yuio.git.RefParser(),
                "<ref>",
                {
                    yuio.git.RefCompleterMode.BRANCH,
                    yuio.git.RefCompleterMode.TAG,
                    yuio.git.RefCompleterMode.HEAD,
                },
            ),
            (
                yuio.git.TagParser(),
                "<tag>",
                {
                    yuio.git.RefCompleterMode.TAG,
                },
            ),
            (
                yuio.git.BranchParser(),
                "<branch>",
                {
                    yuio.git.RefCompleterMode.BRANCH,
                },
            ),
            (
                yuio.git.RemoteParser(),
                "<remote>",
                {
                    yuio.git.RefCompleterMode.REMOTE,
                },
            ),
        ],
    )
    def test_basics(self, parser, name, modes):
        assert not parser.supports_parse_many()
        assert parser.get_nargs() == None
        with pytest.raises(RuntimeError):
            assert parser.parse_many(["HEAD"])
        assert parser.describe() == name
        assert parser.describe_or_def() == name
        assert parser.describe_many() == name
        assert parser.describe_many_or_def() == name
        assert parser.describe_value("HEAD") is None
        assert parser.describe_value_or_def("HEAD") == "HEAD"

        completer = parser.completer()
        assert isinstance(completer, yuio.git.RefCompleter)
        assert completer._modes == modes



# Yuio project, MIT license.
#
# https://github.com/taminomara/yuio/
#
# You're free to copy this file to your project and edit it for your needs,
# just keep this copyright line please :3

"""
This module provides helpers to run subprocesses and get their output.

It handles subprocesses stderr and stdout in a way that doesn't break
loggers from :mod:`yuio.io`.

.. autofunction:: exec

.. autofunction:: sh

.. autoclass:: ExecError

"""

from __future__ import annotations

import logging
import pathlib
import subprocess
import threading

from yuio import _typing as _t

__all__ = [
    "ExecError",
    "exec",
    "sh",
]

_LOGGER = logging.getLogger("yuio.exec")


class ExecError(subprocess.CalledProcessError):
    """
    Raised when executed command returns a non-zero status.

    """

    def __str__(self):
        res = super().__str__()
        if stderr := getattr(self, "stderr"):
            if isinstance(stderr, bytes):
                stderr = stderr.decode(errors="replace")
            res += "\n\nStderr:\n" + stderr
        return res


@_t.overload
def exec(
    *args: str | pathlib.Path,
    cwd: str | pathlib.Path | None = None,
    env: dict[str, str] | None = None,
    input: str | None = None,
    level: int = logging.DEBUG,
    text: _t.Literal[True] = True,
) -> str: ...


@_t.overload
def exec(
    *args: str | pathlib.Path,
    cwd: str | pathlib.Path | None = None,
    env: dict[str, str] | None = None,
    input: bytes | None = None,
    level: int = logging.DEBUG,
    text: _t.Literal[False],
) -> bytes: ...


@_t.overload
def exec(
    *args: str | pathlib.Path,
    cwd: str | pathlib.Path | None = None,
    env: dict[str, str] | None = None,
    input: str | bytes | None = None,
    level: int = logging.DEBUG,
    text: bool,
) -> str | bytes: ...


def exec(
    *args: str | pathlib.Path,
    cwd: None | str | pathlib.Path = None,
    env: dict[str, str] | None = None,
    input: None | str | bytes = None,
    level: int = logging.DEBUG,
    text: bool = True,
):
    """
    Run an executable and return its stdout.

    Command's stderr is interactively printed to the log.

    If the command fails, a :class:`~subprocess.CalledProcessError` is raised.

    :param args:
        command arguments.
    :param cwd:
        set the current directory before the command is executed.
    :param env:
        define the environment variables for the command.
    :param input:
        string with command's stdin.
    :param level:
        logging level for stderr outputs.
        By default, it is set to :data:`logging.DEBUG`, which hides all the output.
    :param text:
        if :data:`True` (default), decode stdout using the system default encoding.
    :return:
        string (or bytes) with command's stdout.

    """

    if cwd is not None:
        if not isinstance(cwd, pathlib.Path):
            cwd = pathlib.Path(cwd)
        cwd = cwd.expanduser().resolve()

    with subprocess.Popen(
        args,
        cwd=cwd,
        env=env,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=text,
        stdin=None if input is None else subprocess.PIPE,
    ) as process:
        stdout = []
        stderr = []

        if _LOGGER.isEnabledFor(level):
            _LOGGER.log(level, " ".join(map(str, args)))

        def read_stderr(fh):
            while True:
                line = fh.readline()
                if not line:
                    return
                if isinstance(line, bytes):
                    line = line.decode(errors="replace")
                stderr.append(line)
                _LOGGER.log(level, line.rstrip("\n"))

        def read_stdout(fh):
            stdout.append(fh.read())

        assert process.stdout
        stdout_thread = threading.Thread(
            target=read_stdout,
            args=(process.stdout,),
            name=f"yuio stdout handler for sub-process",
        )
        stdout_thread.daemon = True
        stdout_thread.start()

        assert process.stderr
        stderr_thread = threading.Thread(
            target=read_stderr,
            args=(process.stderr,),
            name=f"yuio stderr handler for sub-process",
        )
        stderr_thread.daemon = True
        stderr_thread.start()

        if input is not None:
            assert process.stdin is not None
            process.stdin.write(input)
            process.stdin.flush()
            process.stdin.close()

        stdout_thread.join()
        stderr_thread.join()

        process.wait()

        if process.returncode != 0:
            raise ExecError(
                process.returncode, args, output="".join(stdout), stderr="".join(stderr)
            )

        return stdout[0]


@_t.overload
def sh(
    cmd: str,
    /,
    *,
    shell: str = "/bin/sh",
    cwd: str | pathlib.Path | None = None,
    env: dict[str, str] | None = None,
    input: str | None = None,
    level: int = logging.DEBUG,
    text: _t.Literal[True] = True,
) -> str: ...


@_t.overload
def sh(
    cmd: str,
    /,
    *,
    shell: str = "/bin/sh",
    cwd: str | pathlib.Path | None = None,
    env: dict[str, str] | None = None,
    input: bytes | None = None,
    level: int = logging.DEBUG,
    text: _t.Literal[False],
) -> bytes: ...


@_t.overload
def sh(
    cmd: str,
    /,
    *,
    shell: str = "/bin/sh",
    cwd: str | pathlib.Path | None = None,
    env: dict[str, str] | None = None,
    input: bytes | None = None,
    level: int = logging.DEBUG,
    text: bool,
) -> str | bytes: ...


def sh(
    cmd: str,
    /,
    *,
    shell: str = "/bin/sh",
    cwd: str | pathlib.Path | None = None,
    env: dict[str, str] | None = None,
    input: str | bytes | None = None,
    level: int = logging.DEBUG,
    text: bool = True,
):
    """
    Run command in a shell, return its stdout.

    :param cmd:
        shell command.
    :param shell:
        which shell to use. Default is ``/bin/sh``.
    :param cwd:
        set the current directory before the command is executed.
    :param env:
        define the environment variables for the command.
    :param input:
        string with command's stdin.
    :param level:
        logging level for stderr outputs.
        By default, it is set to :data:`logging.DEBUG`, which hides all the output.
    :param text:
        if :data:`True` (default), decode stdout using the system default encoding.
    :return:
        string (or bytes) with command's stdout.

    """

    return exec(
        shell,
        "-c",
        cmd,
        cwd=cwd,
        env=env,
        input=input,
        level=level,
        text=text,
    )
